import os
import typing
from dataclasses import dataclass

import dolfin

from . import Mesh
from . import numpy_mpi
from .dolfin_utils import compute_meshvolume
from .dolfin_utils import get_cavity_volume
from .dolfin_utils import map_vector_field
from .geometry_utils import load_geometry_from_h5
from .geometry_utils import move
from .geometry_utils import save_geometry_to_h5
from .utils import getLogger
from .utils import mpi_comm_world

logger = getLogger(__name__)


@dataclass
class Microstructure:
    """f0 - fibers, s0 - sheets, n0 - cross-sheets"""

    f0: typing.Optional[dolfin.Function] = None
    s0: typing.Optional[dolfin.Function] = None
    n0: typing.Optional[dolfin.Function] = None


@dataclass
class CRLBasis:
    """l0 - longitudinal, c0 - circumferential, r0 - radial"""

    c0: typing.Optional[dolfin.Function] = None
    r0: typing.Optional[dolfin.Function] = None
    l0: typing.Optional[dolfin.Function] = None


@dataclass
class MarkerFunctions:
    """
    # vfun - vertex function
    # efun - edge function
    # ffun - facet function
    # cfun - cell function
    """

    vfun: typing.Optional[dolfin.MeshFunction] = None
    efun: typing.Optional[dolfin.MeshFunction] = None
    ffun: typing.Optional[dolfin.MeshFunction] = None
    cfun: typing.Optional[dolfin.MeshFunction] = None


def get_attribute(obj, key1, key2, default=None):
    f = getattr(obj, key1, None)
    if f is None:
        f = getattr(obj, key2, default)
    return f


class Geometry(object):
    """
    Base class for geometry

    Arguments
    ---------
    mesh : :class:`dolfin.mesh`
        The mesh
    markers : dict
        A dictionary with markers for the mesh (optional)
    marker_functions : :class:`pulse.geometry.MarkerFunctions`
        A Markerfunction object with Meshfunctions for
        the mesh
    microstructure : :class:`pulse.geometry.Microstructure`
        A Markerfunction object with functions for the
        fiber, sheet and sheet normal (optional)
    crl_basis : :class:`pulse.geometry.CRLBasis`
        A CRLBasis objedt with funcions for the circumferential,
        radial and longitudinal vectors. (optional)


    Geometry can be intanciated directly

    Example
    -------

        .. code-block:: python

            import dolfin
            mesh = dolfin.UnitCubeMesh(3,3,3)
            geometry = Geometry(mesh)

    You can load create and instace be loading a geometry from a file

    Example
    -------

        .. code-block:: python

            # Geometry is stored in a file "geometry.h5"
            geo = Geometry.from_file("geometry.h5")

    """

    def __init__(
        self,
        mesh,
        markers=None,
        marker_functions=None,
        microstructure=None,
        crl_basis=None,
    ):
        self.mesh = mesh

        self.markers = markers or {}
        self.marker_functions = marker_functions or MarkerFunctions()
        self.microstructure = microstructure or Microstructure()
        self.crl_basis = crl_basis or CRLBasis()
        self._post_init()

    def _post_init(self):
        pass

    def __repr__(self):
        args = []
        for attr in ("f0", "s0", "n0", "vfun", "ffun", "cfun", "c0", "r0", "l0"):
            if getattr(self, attr) is not None:
                args.append(attr)

        return ("{self.__class__.__name__}" "({self.mesh}), " "{args})").format(
            self=self,
            args=", ".join(args),
        )

    @classmethod
    def from_file(cls, h5name, h5group="", comm=None):
        comm = comm if comm is not None else mpi_comm_world()

        return cls(**cls.load_from_file(h5name, h5group, comm))

    def _copy(self, u=None, factor=1.0):
        """
        Return a copy of this geometry and attach a moved
        mesh according to the provided displacement.
        If no displacement is provided this will just
        return a copy of the original geometry.
        """
        new_mesh = Mesh(self.mesh)

        if u is not None:
            U = move(new_mesh, u, factor)
        else:
            U = u

        marker_functions_ = {}
        for dim, fun in ((0, "vfun"), (1, "efun"), (2, "ffun"), (3, "cfun")):
            f_old = getattr(self, fun)
            if f_old is None:
                continue
            f = dolfin.MeshFunction("size_t", new_mesh, dim, new_mesh.domains())
            f.set_values(f_old.array())
            marker_functions_[fun] = f
        marker_functions = MarkerFunctions(**marker_functions_)

        microstructure_ = {}
        for field in ("f0", "s0", "n0"):
            v0 = getattr(self, field)
            if v0 is None:
                continue
            v = map_vector_field(v0, new_mesh, U)

            microstructure_[field] = v
        microstructure = Microstructure(**microstructure_)

        crl_basis_ = {}
        for basis in ("c0", "r0", "l0"):
            v0 = getattr(self, basis)
            if v0 is None:
                continue
            v = map_vector_field(v0, new_mesh, U)
            crl_basis_[basis] = v
        crl_basis = CRLBasis(**crl_basis_)

        return dict(
            mesh=new_mesh,
            markers=self.markers,
            marker_functions=marker_functions,
            microstructure=microstructure,
            crl_basis=crl_basis,
        )

    def copy(self, u=None, factor=1.0):
        """Return as copy of the geometry with possiblity to
        move the new geometry according to some factor times
        some displacement (u)
        """
        return self.__class__(**self._copy(u=u, factor=factor))

    @staticmethod
    def load_from_file(h5name, h5group, comm):
        logger.debug(f"Load geometry from file {h5name}")

        geo = load_geometry_from_h5(h5name, h5group, include_sheets=True, comm=comm)

        f0 = get_attribute(geo, "f0", "fiber", None)
        s0 = get_attribute(geo, "s0", "sheet", None)
        n0 = get_attribute(geo, "n0", "sheet_normal", None)

        c0 = get_attribute(geo, "c0", "circumferential", None)
        r0 = get_attribute(geo, "r0", "radial", None)
        l0 = get_attribute(geo, "l0", "longitudinal", None)

        vfun = getattr(geo, "vfun", None)
        ffun = getattr(geo, "ffun", None)
        cfun = get_attribute(geo, "cfun", "sfun", None)

        kwargs = {
            "mesh": geo.mesh,
            "markers": geo.markers,
            "marker_functions": MarkerFunctions(vfun=vfun, ffun=ffun, cfun=cfun),
            "microstructure": Microstructure(f0=f0, s0=s0, n0=n0),
            "crl_basis": CRLBasis(c0=c0, r0=r0, l0=l0),
        }

        return kwargs

    def save(
        self,
        h5name,
        h5group="",
        other_functions=None,
        other_attributes=None,
        overwrite_file=False,
        overwrite_group=True,
    ):
        h5name = os.path.splitext(h5name)[0] + ".h5"
        logger.debug(f"Save to {h5name}...")
        save_geometry_to_h5(
            self.mesh,
            h5name=h5name,
            h5group=h5group,
            markers=self.markers or None,
            fields=self.microstructure_list() or None,
            local_basis=self.crl_basis_list() or None,
            meshfunctions=self.meshfunction_list(),
            overwrite_file=overwrite_file,
            overwrite_group=overwrite_group,
        )
        logger.debug("Saved")

    @property
    def top_dim(self):
        """Topological Dimension"""
        return self.mesh.topology().dim()

    @property
    def geo_dim(self):
        """Geometric Dimension"""
        return self.mesh.geometry().dim()

    def crl_basis_list(self):
        """Return a list of the CRL basis in the order
        c0, r0, l0. Basis elements that are none will not
        be included
        """
        if not hasattr(self, "_crl_basis_list"):
            self._crl_basis_list = []
            for basis in ["c0", "r0", "l0"]:
                e = getattr(self.crl_basis, basis)
                if e is not None:
                    self._crl_basis_list.append(e)

        return self._crl_basis_list

    @property
    def nbasis(self):
        return len(self.crl_basis_list())

    @property
    def dx(self):
        """Return the volume measure using self.mesh as domain and
        self.cfun as subdomain_data
        """
        return dolfin.dx(domain=self.mesh, subdomain_data=self.cfun)

    @property
    def ds(self):
        """Return the surface measure of exterior facets using
        self.mesh as domain and self.ffun as subdomain_data
        """
        return dolfin.ds(domain=self.mesh, subdomain_data=self.ffun)

    @property
    def facet_normal(self):
        return dolfin.FacetNormal(self.mesh)

    @property
    def nregions(self):
        return len(self.regions)

    @property
    def regions(self):
        """Return a set of the unique values of the cell function (cfun)"""
        if not hasattr(self, "_regions"):
            try:
                # FIXME
                regions = set(
                    numpy_mpi.gather_broadcast(self.dx.subdomain_data().array()).astype(
                        int,
                    ),
                )
                self._regions = {int(i) for i in regions}
            except AttributeError:
                # self.dx.subdomain_data() is None
                logger.warning("No regions found")
                self._regions = {0}

        return self._regions

    @property
    def meshvols(self):
        """Return a list of the volume of each subdomain defined
        by the cell function (cfun)
        """
        if not hasattr(self, "_meshvols"):
            self._meshvols = [
                compute_meshvolume(dx=self.dx, subdomain_id=int(i))
                for i in self.regions
            ]
        return self._meshvols

    @property
    def meshvol(self):
        """Return the volume of the whole mesh"""
        if not hasattr(self, "_meshvol"):
            self._meshvol = compute_meshvolume(domain=self.mesh)

        return self._meshvol

    def microstructure_list(self):
        """Fibers, sheet and sheet-normals in a list"""
        fields = []
        for field in ["f0", "s0", "n0"]:
            e = getattr(self.microstructure, field)
            if e is not None:
                fields.append(e)
        return fields

    def meshfunction_list(self):
        meshfunctions = {}
        for dim, func in enumerate(["vfun", "efun", "ffun", "cfun"]):
            mf = getattr(self.marker_functions, func)
            if mf is None:
                mf = dolfin.MeshFunction("size_t", self.mesh, dim, self.mesh.domains())
            meshfunctions[dim] = mf
        return meshfunctions

    @property
    def vfun(self):
        """Vertex Function"""
        return self.marker_functions.vfun

    @property
    def efun(self):
        """Edge Function"""
        return self.marker_functions.efun

    @property
    def ffun(self):
        """Facet Function"""
        return self.marker_functions.ffun

    @property
    def cfun(self):
        """Cell Function"""
        return self.marker_functions.cfun

    @property
    def f0(self):
        """Fibers"""
        return self.microstructure.f0

    @property
    def s0(self):
        """Sheets"""
        return self.microstructure.s0

    @property
    def n0(self):
        """Cross-Sheets"""
        return self.microstructure.n0

    @property
    def l0(self):
        """Longitudinal"""
        return self.crl_basis.l0

    @property
    def c0(self):
        """Circumferential"""
        return self.crl_basis.c0

    @property
    def r0(self):
        """Radial"""
        return self.crl_basis.r0


class HeartGeometry(Geometry):
    def _post_init(self):
        self.xshift = 0.0

    @property
    def is_biv(self):
        if "ENDO_RV" in self.markers:
            # FIXME
            return self.markers["ENDO_RV"][0] in set(
                numpy_mpi.gather_broadcast(self.ffun.array()),
            )

        return False

    def cavity_volume(self, chamber="lv", u=None):
        return get_cavity_volume(self, chamber=chamber, u=u, xshift=self.xshift)

    @property
    def base_mean_position(self):
        """
        Return mean coordinates of the base
        (serial only?)
        """
        import numpy as np

        (facet_indices,) = np.where(self.ffun.array() == self.markers["BASE"][0])
        point_inidces = []
        for facet in dolfin.facets(self.mesh):
            if facet.index() in facet_indices:
                point_inidces.extend(facet.entities(0).tolist())

        return np.mean(self.mesh.coordinates()[point_inidces, :], 0)

    def update_xshift(self):
        self.xshift = self.base_mean_position[0]



import sys
import time
from pathlib import Path

import dolfin

from . import as_backend_type
from . import assemble
from .utils import enlist
from .utils import getLogger
from .utils import mpi_comm_world

logger = getLogger(__name__)


def dump_matrix_to_mtx(A: dolfin.PETScMatrix, filename: Path) -> None:
    with open(filename, "w") as f:
        mat = as_backend_type(A).mat()
        (num_rows, num_columns) = mat.size
        (ai, aj, av) = mat.getValuesCSR()
        num_nonzeros = len(av)
        f.write("%%MatrixMarket matrix coordinate real general\n")
        f.write(f"% Generated by {' '.join(sys.argv)}\n")
        f.write(f"{num_rows} {num_columns} {num_nonzeros}\n")
        for i in range(num_rows):
            for k in range(ai[i], ai[i + 1]):
                f.write(f"{i + 1} {aj[k] + 1} {av[k]}\n")


class NonlinearProblem(dolfin.NonlinearProblem):
    def __init__(
        self,
        J,
        F,
        bcs,
        output_matrix=False,
        output_matrix_path="output",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._J = J
        self._F = F

        self.bcs = enlist(bcs)
        self.output_matrix = output_matrix
        self.output_matrix_path = output_matrix_path
        self.verbose = True
        self.n = 0

    def F(self, b: dolfin.PETScVector, x: dolfin.PETScVector):
        assemble(self._F, tensor=b)
        for bc in self.bcs:
            bc.apply(b, x)

    def J(self, A: dolfin.PETScMatrix, x: dolfin.PETScVector):
        assemble(self._J, tensor=A)
        for bc in self.bcs:
            bc.apply(A)

        if self.output_matrix:
            filename = Path("{self.output_matrix_path}/J_{self.n:04d}.mtx")
            filename.parent.mkdir(exist_ok=True, parents=True)
            dump_matrix_to_mtx(A, filename)
            if self.verbose:
                print(f"Assembled matrix written to {filename}")
            self.n = self.n + 1


class NonlinearSolver:
    def __init__(
        self,
        problem: NonlinearProblem,
        state,
        parameters=None,
    ):
        dolfin.PETScOptions.clear()
        self.update_parameters(parameters)
        self._problem = problem
        self._state = state

        self._solver = dolfin.PETScSNESSolver(mpi_comm_world())
        self._solver.set_from_options()

        self._solver.parameters.update(self.parameters)
        self._snes = self._solver.snes()
        self._snes.setConvergenceHistory()

        logger.debug(f"Linear Solver : {self._solver.parameters['linear_solver']}")
        logger.debug(f"Preconditioner:  {self._solver.parameters['preconditioner']}")
        logger.debug(f"atol: {self._solver.parameters['absolute_tolerance']}")
        logger.debug(f"rtol: {self._solver.parameters['relative_tolerance']}")
        logger.debug(f" Size          : {self._state.function_space().dim()}")
        dolfin.PETScOptions.clear()

    def update_parameters(self, parameters):
        ps = NonlinearSolver.default_solver_parameters()
        if hasattr(self, "parameters"):
            ps.update(self.parameters)
        if parameters is not None:
            ps.update(parameters)
        petsc = ps.pop("petsc")

        for k, v in petsc.items():
            if v is not None:
                dolfin.PETScOptions.set(k, v)
        self.verbose = ps.pop("verbose", False)
        if self.verbose:
            dolfin.PETScOptions.set("ksp_monitor")
            dolfin.PETScOptions.set("log_view")
            dolfin.PETScOptions.set("ksp_view")
            dolfin.PETScOptions.set("pc_view")
            dolfin.PETScOptions.set("mat_superlu_dist_statprint", True)
            ps["lu_solver"]["report"] = True
            ps["lu_solver"]["verbose"] = True
            ps["report"] = True
            ps["krylov_solver"]["monitor_convergence"] = True
        self.parameters = ps

    @staticmethod
    def default_solver_parameters():
        return {
            "petsc": {
                "ksp_type": "preonly",
                "pc_type": "lu",
                "pc_factor_mat_solver_type": "mumps",
                "mat_mumps_icntl_33": 0,
            },
            "verbose": False,
            "linear_solver": "mumps",
            "preconditioner": "lu",
            "error_on_nonconvergence": False,
            "relative_tolerance": 1e-5,
            "absolute_tolerance": 1e-5,
            "maximum_iterations": 20,
            "report": False,
            "krylov_solver": {
                "absolute_tolerance": 1e-13,
                "relative_tolerance": 1e-13,
                "maximum_iterations": 1000,
                "monitor_convergence": False,
            },
            "lu_solver": {"report": False, "symmetric": False, "verbose": False},
        }

    def solve(self):
        """Solve the problem.

        Returns
        -------
        residual : _solver.snes (???)
            A measure of the accuracy (convergence and error)
            of the performed computation.
        """

        logger.debug(" Solving NonLinearProblem ...")

        start = time.time()
        self._solver.solve(self._problem, self._state.vector())
        end = time.time()

        logger.debug(f" ... Done in {end - start:.3f} s")

        residuals = self._snes.getConvergenceHistory()[0]
        num_iterations = self._snes.getLinearSolveIterations()
        logger.debug(f"Iterations    : {num_iterations}")
        if num_iterations > 0:
            logger.debug(f"Resiudal      : {residuals[-1]}")

        return num_iterations, self._snes.converged



import itertools

import numpy as np
import pytest

from pulse.iterate import iterate

try:
    import dolfin_adjoint

    has_dolfin_adjoint = True
except ImportError:
    has_dolfin_adjoint = False

import dolfin
from utils import make_lv_mechanics_problem

has_dolfin_adjoint = False
if has_dolfin_adjoint:
    # Run tests with and without annotation
    annotates = (False, True)
else:
    annotates = (False, False)


cases = itertools.product((False, True), annotates)


@pytest.fixture
def problem():
    problem = make_lv_mechanics_problem("R_0")
    return problem


def test_iterate_pressure(problem):
    target_pressure = 0.1
    plv = [p.traction for p in problem.bcs.neumann if p.name == "lv"]
    pressure = plv[0]

    if has_dolfin_adjoint:
        dolfin.parameters["adjoint"]["stop_annotating"] = False
        dolfin_adjoint.adj_reset()

    iterate(problem, pressure, target_pressure)

    if has_dolfin_adjoint:
        # Check the recording
        assert dolfin_adjoint.replay_dolfin(tol=1e-12)

    # Check that the pressure is correct
    assert float(plv[0]) == target_pressure
    # Check that the state is nonzero
    assert dolfin.norm(problem.state.vector()) > 0


def test_iterate_gamma(problem):
    target_gamma = 0.001
    gamma = problem.material.activation

    if has_dolfin_adjoint:
        dolfin.parameters["adjoint"]["stop_annotating"] = False
        dolfin_adjoint.adj_reset()

    iterate(problem, gamma, target_gamma)

    assert all(gamma.vector().get_local() == target_gamma)
    assert dolfin.norm(problem.state.vector()) > 0

    if has_dolfin_adjoint:
        assert dolfin_adjoint.replay_dolfin(tol=1e-12)


def test_iterate_gamma_regional():
    problem = make_lv_mechanics_problem("regional")
    target_gamma = problem.material.activation.vector().get_local()
    for i in range(len(target_gamma)):
        target_gamma[i] = 0.01 - i * 0.001
    print(target_gamma)

    gamma = problem.material.activation

    if has_dolfin_adjoint:
        dolfin.parameters["adjoint"]["stop_annotating"] = False
        dolfin_adjoint.adj_reset()

    iterate(problem, gamma, target_gamma)

    print(gamma.vector().get_local())
    assert np.all(gamma.vector().get_local() - target_gamma < 1e-12)
    assert dolfin.norm(problem.state.vector()) > 0

    if has_dolfin_adjoint:
        assert dolfin_adjoint.replay_dolfin(tol=1e-12)


@pytest.mark.parametrize("continuation", [True, False])
def test_iterate_gamma_cg1(continuation):
    problem = make_lv_mechanics_problem("CG_1")
    V = problem.material.activation.function_space()
    target_gamma = dolfin.interpolate(dolfin.Expression("0.01 * x[0]", degree=1), V)
    gamma = problem.material.activation

    if has_dolfin_adjoint:
        dolfin.parameters["adjoint"]["stop_annotating"] = False
        dolfin_adjoint.adj_reset()

    iterate(problem, gamma, target_gamma, continuation=continuation)

    assert np.all(
        gamma.vector().get_local() - target_gamma.vector().get_local() < 1e-12,
    )
    assert dolfin.norm(problem.state.vector()) > 0

    if has_dolfin_adjoint:
        assert dolfin_adjoint.replay_dolfin(tol=1e-12)


def test_iterate_gamma_pressure(problem):
    target_pressure = 1.0
    plv = [p.traction for p in problem.bcs.neumann if p.name == "lv"]
    pressure = plv[0]

    target_gamma = 0.01
    gamma = problem.material.activation

    if has_dolfin_adjoint:
        dolfin.parameters["adjoint"]["stop_annotating"] = False
        dolfin_adjoint.adj_reset()

    iterate(problem, (pressure, gamma), (target_pressure, target_gamma))

    assert all(gamma.vector().get_local() == target_gamma)
    assert float(plv[0]) == target_pressure
    assert dolfin.norm(problem.state.vector()) > 0

    if has_dolfin_adjoint:
        assert dolfin_adjoint.replay_dolfin(tol=1e-12)


def test_iterate_regional_gamma_pressure():
    problem = make_lv_mechanics_problem("regional")
    target_gamma = problem.material.activation.vector().get_local()
    for i in range(len(target_gamma)):
        target_gamma[i] = 0.01 - i * 0.001
    gamma = problem.material.activation

    target_pressure = 0.1
    plv = [p.traction for p in problem.bcs.neumann if p.name == "lv"]
    pressure = plv[0]

    iterate(problem, (pressure, gamma), (target_pressure, target_gamma))


def test_iterate_zero():
    problem = make_lv_mechanics_problem("R_0")
    iterate(problem, problem.material.activation, 1e-5)



import logging as _logging
import os
from importlib.metadata import metadata

from dolfin import as_backend_type
from dolfin import assemble
from dolfin import Constant
from dolfin import DirichletBC
from dolfin import Function
from dolfin import FunctionAssigner
from dolfin import interpolate
from dolfin import Mesh
from dolfin import project


meta = metadata("fenics-pulse")
__version__ = meta["Version"]
__author__ = meta["Author-email"]
__license__ = meta["License"]
__email__ = meta["Author-email"]
__program_name__ = meta["Name"]


Constants = Constant
Functions = Function
# Check if dolfin-adjoint is installed, and if
# the 'DOLFIN_ADJOINT' flag is set to 0, then
# we remove it from sys.modules
try:
    import dolfin_adjoint  # noqa: F401
except ImportError:
    has_dolfin_adjoint = False
else:
    has_dolfin_adjoint = True
    if not bool(int(os.getenv("DOLFIN_ADJOINT", "1"))):
        import sys

        sys.modules["dolfin_adjoint"] = None  # type: ignore
        _logging.warning("Dolfin-adjoint found but will be turned off")
    else:
        import dolfin as _dolfin
        from dolfin_adjoint import (  # noqa: F811
            as_backend_type,  # noqa: F811
            Constant,  # noqa: F811
            Function,  # noqa: F811
            FunctionAssigner,  # noqa: F811
            assemble,  # noqa: F811
            interpolate,  # noqa: F811
            project,  # noqa: F811
            Mesh,  # noqa: F811
            DirichletBC,  # noqa: F811
        )

        Constants = (_dolfin.Constant, Constant)
        Functions = (_dolfin.Function, Function)


import daiquiri as _daiquiri


from . import dolfin_utils
from . import geometry
from . import geometry_utils
from . import io_utils
from . import iterate
from . import kinematics
from . import material
from . import mechanicsproblem
from . import numpy_mpi
from . import solver
from . import unloader
from . import utils
from .dolfin_utils import MixedParameter
from .dolfin_utils import QuadratureSpace
from .dolfin_utils import RegionalParameter
from .example_meshes import mesh_paths
from .geometry import CRLBasis
from .geometry import Geometry
from .geometry import HeartGeometry
from .geometry import MarkerFunctions
from .geometry import Microstructure
from .kinematics import DeformationGradient
from .kinematics import EulerAlmansiStrain
from .kinematics import GreenLagrangeStrain
from .kinematics import InversePiolaTransform
from .kinematics import Jacobian
from .kinematics import LeftCauchyGreen
from .kinematics import PiolaTransform
from .kinematics import RightCauchyGreen
from .kinematics import SecondOrderIdentity
from .material import Guccione
from .material import HolzapfelOgden
from .material import LinearElastic
from .material import Material
from .material import NeoHookean
from .material import StVenantKirchhoff
from .material import ActiveModel
from .material import ActiveModels
from .mechanicsproblem import BoundaryConditions
from .mechanicsproblem import MechanicsProblem
from .mechanicsproblem import NeumannBC
from .mechanicsproblem import RobinBC
from .solver import NonlinearProblem
from .solver import NonlinearSolver
from .unloader import FixedPointUnloader
from .unloader import MeshUnloader
from .unloader import RaghavanUnloader
from .utils import annotation


def init_ffc():
    import dolfin as _dolfin

    flags = ["-O3", "-ffast-math"]  # , "-march=native"]
    _dolfin.parameters["form_compiler"]["quadrature_degree"] = 4
    _dolfin.parameters["form_compiler"]["representation"] = "uflacs"
    _dolfin.parameters["form_compiler"]["cpp_optimize"] = True
    _dolfin.parameters["form_compiler"]["cpp_optimize_flags"] = " ".join(flags)
    _dolfin.set_log_level(_logging.WARNING)


init_ffc()


def set_log_level(level):
    from daiquiri import set_default_log_levels

    loggers = [
        "pulse.utils.logger",
        "pulse.dolfin_utils.logger",
        "pulse.io_utils.logger",
        "pulse.solver.logger",
        "pulse.mechanicsproblem.logger",
        "pulse.iterate.logger",
        "pulse.unloader.logger",
        "pulse.geometry.logger",
        "pulse.geometry_utils.logger",
    ]
    set_default_log_levels((logger, level) for logger in loggers)


_daiquiri.setup(level=_logging.INFO)


ffc_logger = _logging.getLogger("FFC")
ffc_logger.setLevel(_logging.WARNING)
ffc_logger.addFilter(utils.mpi_filt)

ufl_logger = _logging.getLogger("UFL")
ufl_logger.setLevel(_logging.FATAL)
ufl_logger.addFilter(utils.mpi_filt)


__all__ = [
    "annotation",
    "utils",
    "dolfin_utils",
    "io_utils",
    "numpy_mpi",
    "kinematics",
    "mechanicsproblem",
    "iterate",
    "unloader",
    "Geometry",
    "CRLBasis",
    "HeartGeometry",
    "Microstructure",
    "MarkerFunctions",
    "QuadratureSpace",
    "MixedParameter",
    "RegionalParameter",
    "material",
    "MeshUnloader",
    "RaghavanUnloader",
    "FixedPointUnloader",
    "BoundaryConditions",
    "MechanicsProblem",
    "NeumannBC",
    "RobinBC",
    "Material",
    "HolzapfelOgden",
    "Guccione",
    "LinearElastic",
    "NeoHookean",
    "StVenantKirchhoff",
    "SecondOrderIdentity",
    "DeformationGradient",
    "Jacobian",
    "GreenLagrangeStrain",
    "LeftCauchyGreen",
    "RightCauchyGreen",
    "EulerAlmansiStrain",
    "PiolaTransform",
    "InversePiolaTransform",
    "set_log_level",
    "solver",
    "mesh_paths",
    "NonlinearProblem",
    "NonlinearSolver",
    "geometry",
    "geometry_utils",
    "ActiveModel",
    "ActiveModels",
    "Function",
    "interpolate",
    "Constant",
    "project",
    "assemble",
    "FunctionAssigner",
    "Mesh",
    "DirichletBC",
    "as_backend_type",
]



#!/usr/bin/env python
import logging

import dolfin
import h5py
import numpy as np

from . import numpy_mpi
from .dolfin_utils import get_pressure
from .iterate import iterate
from .iterate import logger as logger_it
from .utils import annotation
from .utils import getLogger
from .utils import mpi_comm_world

logger = getLogger(__name__)


class ResidualCalculator(object):
    def __init__(self, mesh):
        try:
            from scipy import spatial

        except ImportError:
            logger.warning(
                (
                    "Scipy is not install. Residual in the unloading "
                    "algorithm cannot be computed. Please install "
                    'scipy "pip install scipy" if you want to compute '
                    "the residual"
                ),
            )
            self.bbtree = None

        else:
            self.mesh = mesh
            d = self.mesh.topology().dim()
            local_points = [v.point() for v in dolfin.vertices(self.mesh)]
            coords = [(p.x(), p.y(), p.z()) for p in local_points]

            # FIXME
            coords = numpy_mpi.gather_broadcast(np.array(coords).flatten())
            coords.resize(int(len(coords) / d), d)

            self.bbtree = spatial.KDTree(coords)

    def calculate_residual(self, mesh2):
        boundmesh = dolfin.BoundaryMesh(mesh2, "exterior")
        d = max(
            [
                self.bbtree.query(dolfin.Vertex(boundmesh, v_idx).point().array())[0]
                for v_idx in range(boundmesh.num_vertices())
            ],
        )

        return dolfin.MPI.max(mpi_comm_world(), d)


def inflate_to_pressure(target_pressure, problem, ntries=5, n=2, annotate=False):
    logger.debug(f"\nInflate geometry to p = {target_pressure} kPa")
    pressure = get_pressure(problem)
    solve(target_pressure, problem, pressure, ntries, n, annotate)

    return problem.get_displacement(annotate=annotate)


def print_volumes(geometry, logger=logger, txt="original", u=None):
    logger.info(
        ("\nLV Volume of {} geometry = {:.3f} ml" "").format(
            txt,
            geometry.cavity_volume(chamber="lv"),
        ),
    )
    if geometry.is_biv:
        logger.info(
            ("RV Volume of {} geometry = {:.3f} ml" "").format(
                txt,
                geometry.cavity_volume(chamber="rv"),
            ),
        )


def solve(target_pressure, problem, pressure, ntries=5, n=2, annotate=False):
    annotation.annotate = False

    level = logger_it.logger.level
    logger_it.setLevel(logging.WARNING)

    iterate(problem, pressure, target_pressure)

    annotation.annotate = True
    # Only record the last solve, otherwise it becomes too
    # expensive memorywise.
    problem.solve()

    logger_it.setLevel(level)
    w = problem.state.copy(deepcopy=True)
    return w


def load_opt_target(h5name, h5group, key="volume", data="simulated"):
    with h5py.File(h5name) as f:
        vols = [a[:][0] for a in f[h5group]["passive_inflation"][key][data].values()]

    return vols


def continuation_step(params, it_, paramvec):
    # Use data from the two prevoious steps and continuation
    # to get a good next gues
    values = []
    vols = []

    v_target = load_opt_target(params["sim_file"], "0", "volume", "target")
    for it in range(it_):
        p_tmp = dolfin.Function(paramvec.function_space())
        load_material_parameter(params["sim_file"], str(it), p_tmp)

        values.append(numpy_mpi.gather_vector(p_tmp.vector()))

        v = load_opt_target(params["sim_file"], str(it), "volume", "simulated")
        vols.append(v)

    ed_vols = np.array(vols).T[-1]
    # Make continuation approximation
    delta = (v_target[-1] - ed_vols[-2]) / (ed_vols[-1] - ed_vols[-2])
    a_cont = (1 - delta) * values[-2] + delta * values[-1]
    a_prev = values[-1]

    # Make sure next step is not to far away
    if hasattr(a_cont, "__len__"):
        a_next = np.array(
            [
                min(max(a_cont[i], a_prev[i] / 2), a_prev[i] * 2)
                for i in range(len(a_cont))
            ],
        )

        # Just make sure that we are within the given bounds
        a = np.array(
            [
                min(
                    max(a_next[i], params["Optimization_parameters"]["matparams_min"]),
                    params["Optimization_parameters"]["matparams_max"],
                )
                for i in range(len(a_cont))
            ],
        )

    else:
        a_next = min(max(a_cont, a_prev / 2), a_prev * 2)

        # Just make sure that we are within the given bounds
        a = min(
            max(a_next, params["Optimization_parameters"]["matparams_min"]),
            params["Optimization_parameters"]["matparams_max"],
        )

    numpy_mpi.assign_to_vector(paramvec.vector(), a)


def load_material_parameter(h5name, h5group, paramvec):
    logger.info(f"Load {h5name}:{h5group}")
    group = "/".join([h5group, "passive_inflation", "optimal_control"])
    with dolfin.HDF5File(mpi_comm_world(), h5name, "r") as h5file:
        h5file.read(paramvec, group)



import dolfin
import numpy as np
import pytest

try:
    from dolfin_adjoint import UnitCubeMesh, interpolate
except ImportError:
    from dolfin import UnitCubeMesh, interpolate

from pulse.example_meshes import mesh_paths
from pulse.geometry import CRLBasis
from pulse.geometry import Geometry
from pulse.geometry import HeartGeometry
from pulse.geometry import MarkerFunctions
from pulse.geometry import Microstructure


class Free(dolfin.SubDomain):
    def inside(self, x, on_boundary):
        return x[0] > (1.0 - dolfin.DOLFIN_EPS) and on_boundary


class Fixed(dolfin.SubDomain):
    def inside(self, x, on_boundary):
        return x[0] < dolfin.DOLFIN_EPS and on_boundary


fixed = Fixed()
fixed_marker = 1

free = Free()
free_marker = 2


def strain_markers_3d(mesh, nregions):
    strain_markers = dolfin.MeshFunction("size_t", mesh, 3)
    strain_markers.set_all(0)
    xs = np.linspace(0, 1, nregions + 1)

    region = 0
    for it_x in range(nregions):
        for it_y in range(nregions):
            for it_z in range(nregions):
                region += 1
                domain_str = ""

                domain_str += f"x[0] >= {xs[it_x]}"
                domain_str += f" && x[1] >= {xs[it_y]}"
                domain_str += f" && x[2] >= {xs[it_z]}"
                domain_str += f" && x[0] <= {xs[it_x + 1]}"
                domain_str += f" && x[1] <= {xs[it_y + 1]}"
                domain_str += f" && x[2] <= {xs[it_z + 1]}"

                len_sub = dolfin.CompiledSubDomain(domain_str)
                len_sub.mark(strain_markers, region)
    return strain_markers


@pytest.fixture
def unitcube_geometry():
    N = 2
    mesh = UnitCubeMesh(N, N, N)

    V_f = dolfin.VectorFunctionSpace(mesh, "CG", 1)

    l0 = interpolate(dolfin.Expression(("1.0", "0.0", "0.0"), degree=1), V_f)
    r0 = interpolate(dolfin.Expression(("0.0", "1.0", "0.0"), degree=1), V_f)
    c0 = interpolate(dolfin.Expression(("0.0", "0.0", "1.0"), degree=1), V_f)

    crl_basis = CRLBasis(l0=l0, r0=r0, c0=c0)

    cfun = strain_markers_3d(mesh, 2)

    ffun = dolfin.MeshFunction("size_t", mesh, 2)
    ffun.set_all(0)
    fixed.mark(ffun, fixed_marker)
    free.mark(ffun, free_marker)

    marker_functions = MarkerFunctions(ffun=ffun, cfun=cfun)

    # Fibers
    f0 = interpolate(dolfin.Expression(("1.0", "0.0", "0.0"), degree=1), V_f)
    s0 = interpolate(dolfin.Expression(("0.0", "1.0", "0.0"), degree=1), V_f)
    n0 = interpolate(dolfin.Expression(("0.0", "0.0", "1.0"), degree=1), V_f)

    microstructure = Microstructure(f0=f0, s0=s0, n0=n0)

    geometry = Geometry(
        mesh=mesh,
        marker_functions=marker_functions,
        microstructure=microstructure,
        crl_basis=crl_basis,
    )

    return geometry


def test_create_geometry(unitcube_geometry):
    # Check that attributes are set correctly
    assert unitcube_geometry.f0 is not None
    assert unitcube_geometry.s0 is not None
    assert unitcube_geometry.n0 is not None
    assert unitcube_geometry.microstructure is not None

    assert unitcube_geometry.l0 is not None
    assert unitcube_geometry.c0 is not None
    assert unitcube_geometry.r0 is not None
    assert unitcube_geometry.crl_basis is not None

    assert unitcube_geometry.ffun is not None
    assert unitcube_geometry.cfun is not None

    assert unitcube_geometry.mesh is not None
    assert unitcube_geometry.markers is not None


def test_mesvolumes(unitcube_geometry):
    assert (float(unitcube_geometry.meshvol) - 1.0) < dolfin.DOLFIN_EPS_LARGE
    assert all(
        [
            (float(f) - 0.125) < dolfin.DOLFIN_EPS_LARGE
            for f in unitcube_geometry.meshvols
        ],
    )
    assert unitcube_geometry.nregions == 8


def test_crl_basis(unitcube_geometry):
    assert unitcube_geometry.crl_basis_list
    assert unitcube_geometry.nbasis == 3


def test_simple_ellipsoid():
    geometry = HeartGeometry.from_file(mesh_paths["simple_ellipsoid"])

    assert abs(geometry.cavity_volume() - 0.7494375870292063) < 1e-14
    assert geometry.is_biv is False


def test_biv_ellipsoid():
    geometry = HeartGeometry.from_file(mesh_paths["biv_ellipsoid"])

    assert abs(geometry.cavity_volume(chamber="lv") - 0.7415450456459624) < 1e-14
    assert abs(geometry.cavity_volume(chamber="rv") - 0.8141695718969176) < 1e-14
    assert geometry.is_biv is True


if __name__ == "__main__":
    # geo = unitcube_geometry()
    # test_create_geometry(geo)
    # test_mesvolumes(geo)
    test_simple_ellipsoid()
    test_biv_ellipsoid()



import typing
from dataclasses import dataclass
from functools import partial

import dolfin

try:
    import ufl_legacy as ufl
except ImportError:
    import ufl

from . import Constant
from . import DirichletBC
from . import Function
from . import FunctionAssigner
from . import has_dolfin_adjoint
from . import kinematics
from .dolfin_utils import list_sum
from .geometry import Geometry
from .geometry import HeartGeometry
from .material import Material
from .solver import NonlinearProblem
from .solver import NonlinearSolver
from .utils import get_lv_marker
from .utils import getLogger

logger = getLogger(__name__)


@dataclass
class NeumannBC:
    traction: typing.Union[float, ufl.Coefficient]
    marker: int
    name: str = ""


@dataclass
class RobinBC:
    value: typing.Union[float, ufl.Coefficient]
    marker: int


dirichlet_types = typing.Union[
    typing.Callable[[dolfin.FunctionSpace], dolfin.DirichletBC],
    dolfin.DirichletBC,
]


@dataclass
class BoundaryConditions:
    neumann: typing.Sequence[NeumannBC] = ()
    dirichlet: typing.Sequence[dirichlet_types] = ()
    robin: typing.Sequence[RobinBC] = ()
    body_force: typing.Sequence[ufl.Coefficient] = ()


def dirichlet_fix_base(W, ffun, marker):
    """Fix the basal plane."""
    V = W if W.sub(0).num_sub_spaces() == 0 else W.sub(0)
    bc = DirichletBC(V, Constant((0, 0, 0)), ffun, marker)
    return bc


def dirichlet_fix_base_directional(W, ffun, marker, direction=0):
    V = W if W.sub(0).num_sub_spaces() == 0 else W.sub(0)
    bc = DirichletBC(V.sub(direction), Constant(0.0), ffun, marker)
    return bc


def cardiac_boundary_conditions(
    geometry,
    pericardium_spring=0.0,
    base_spring=0.0,
    base_bc="fix_x",
    **kwargs,
):
    msg = (
        "Cardiac boundary conditions can only be applied to a "
        "HeartGeometry got {}".format(type(geometry))
    )
    assert isinstance(geometry, HeartGeometry), msg

    # Neumann BC
    lv_marker = get_lv_marker(geometry)
    lv_pressure = NeumannBC(
        traction=Constant(0.0, name="lv_pressure"),
        marker=lv_marker,
        name="lv",
    )
    neumann_bc = [lv_pressure]

    if geometry.is_biv:
        rv_pressure = NeumannBC(
            traction=Constant(0.0, name="rv_pressure"),
            marker=geometry.markers["ENDO_RV"][0],
            name="rv",
        )

        neumann_bc += [rv_pressure]

    # Robin BC
    if pericardium_spring > 0.0:
        robin_bc = [
            RobinBC(
                value=Constant(pericardium_spring),
                marker=geometry.markers["EPI"][0],
            ),
        ]

    else:
        robin_bc = []

    # Apply a linear sprint robin type BC to limit motion
    if base_spring > 0.0:
        robin_bc += [
            RobinBC(value=Constant(base_spring), marker=geometry.markers["BASE"][0]),
        ]

    # Dirichlet BC
    if base_bc == "fixed":
        dirichlet_bc = [
            partial(
                dirichlet_fix_base,
                ffun=geometry.ffun,
                marker=geometry.markers["BASE"][0],
            ),
        ]

    elif base_bc == "fix_x":
        dirichlet_bc = [
            partial(
                dirichlet_fix_base_directional,
                ffun=geometry.ffun,
                marker=geometry.markers["BASE"][0],
            ),
        ]
    else:
        raise ValueError(f"Unknown base bc {base_bc}")

    boundary_conditions = BoundaryConditions(
        dirichlet=dirichlet_bc,
        neumann=neumann_bc,
        robin=robin_bc,
    )

    return boundary_conditions


class SolverDidNotConverge(Exception):
    pass


class MechanicsProblem(object):
    """
    Base class for mechanics problem
    """

    def __init__(
        self,
        geometry: Geometry,
        material: Material,
        bcs=None,
        bcs_parameters=None,
        solver_parameters=None,
    ):
        logger.debug("Initialize mechanics problem")
        self.geometry = geometry
        self.material = material

        self._handle_bcs(bcs=bcs, bcs_parameters=bcs_parameters)

        # Make sure that the material has microstructure information
        for attr in ("f0", "s0", "n0"):
            setattr(self.material, attr, getattr(self.geometry, attr))

        self.solver_parameters = NonlinearSolver.default_solver_parameters()
        if solver_parameters is not None:
            self.solver_parameters.update(**solver_parameters)

        self._init_spaces()
        self._init_forms()

    def _handle_bcs(self, bcs, bcs_parameters):
        if bcs is None:
            if isinstance(self.geometry, HeartGeometry):
                self.bcs_parameters = MechanicsProblem.default_bcs_parameters()
                if bcs_parameters is not None:
                    self.bcs_parameters.update(**bcs_parameters)
            else:
                raise ValueError(
                    ("Please provide boundary conditions " "to MechanicsProblem"),
                )

            self.bcs = cardiac_boundary_conditions(self.geometry, **self.bcs_parameters)

        else:
            self.bcs = bcs

            # TODO: FIX THIS or require this
            # Just store this as well in case both is provided
            self.bcs_parameters = MechanicsProblem.default_bcs_parameters()
            if bcs_parameters is not None:
                self.bcs_parameters.update(**bcs_parameters)

    @staticmethod
    def default_bcs_parameters():
        return dict(pericardium_spring=0.0, base_spring=0.0, base_bc="fixed")

    def _init_spaces(self):
        logger.debug("Initialize spaces for mechanics problem")
        mesh = self.geometry.mesh

        P2 = dolfin.VectorElement("Lagrange", mesh.ufl_cell(), 2)
        P1 = dolfin.FiniteElement("Lagrange", mesh.ufl_cell(), 1)

        # P2_space = FunctionSpace(mesh, P2)
        # P1_space = FunctionSpace(mesh, P1)
        self.state_space = dolfin.FunctionSpace(mesh, P2 * P1)

        self.state = Function(self.state_space, name="state")
        self.state_test = dolfin.TestFunction(self.state_space)

    def _init_forms(self):
        logger.debug("Initialize forms mechanics problem")
        # Displacement and hydrostatic_pressure
        u, p = dolfin.split(self.state)
        v, q = dolfin.split(self.state_test)

        # Some mechanical quantities
        F = dolfin.variable(kinematics.DeformationGradient(u))
        J = kinematics.Jacobian(F)
        dx = self.geometry.dx

        internal_energy = self.material.strain_energy(
            F,
        ) + self.material.compressibility(p, J)

        self._virtual_work = dolfin.derivative(
            internal_energy * dx,
            self.state,
            self.state_test,
        )

        external_work = self._external_work(u, v)
        if external_work is not None:
            self._virtual_work += external_work

        self._set_dirichlet_bc()
        self._jacobian = dolfin.derivative(
            self._virtual_work,
            self.state,
            dolfin.TrialFunction(self.state_space),
        )
        self._init_solver()

    def _init_solver(self):
        if has_dolfin_adjoint:
            from dolfin_adjoint import (
                NonlinearVariationalProblem,
                NonlinearVariationalSolver,
            )

            self._problem = NonlinearVariationalProblem(
                J=self._jacobian,
                F=self._virtual_work,
                u=self.state,
                bcs=self._dirichlet_bc,
            )
            self.solver = NonlinearVariationalSolver(self._problem)
        else:
            self._problem = NonlinearProblem(
                J=self._jacobian,
                F=self._virtual_work,
                bcs=self._dirichlet_bc,
            )
            self.solver = NonlinearSolver(
                self._problem,
                self.state,
                parameters=self.solver_parameters,
            )

    def _set_dirichlet_bc(self):
        # DirichletBC
        for dirichlet_bc in self.bcs.dirichlet:
            msg = (
                "DirichletBC only implemented for as "
                "callable. Please provide DirichletBC "
                "as a callable with argument being the "
                "state space only"
            )

            if hasattr(dirichlet_bc, "__call__"):
                try:
                    self._dirichlet_bc = dirichlet_bc(self.state_space)
                except Exception as ex:
                    logger.error(msg)
                    raise ex
            else:
                raise NotImplementedError(msg)

    def _external_work(self, u, v):
        F = dolfin.variable(kinematics.DeformationGradient(u))

        N = self.geometry.facet_normal
        ds = self.geometry.ds
        dx = self.geometry.dx

        external_work = []

        for neumann in self.bcs.neumann:
            n = neumann.traction * ufl.cofac(F) * N
            external_work.append(dolfin.inner(v, n) * ds(neumann.marker))

        for robin in self.bcs.robin:
            external_work.append(dolfin.inner(robin.value * u, v) * ds(robin.marker))

        for body_force in self.bcs.body_force:
            external_work.append(
                -dolfin.derivative(dolfin.inner(body_force, u) * dx, u, v),
            )

        if len(external_work) > 0:
            return list_sum(external_work)

        return None

    def reinit(self, state, annotate=False):
        """Reinitialze state"""

        if has_dolfin_adjoint:
            try:
                self.state.assign(state, annotate=annotate)
            except Exception as ex:
                print(ex)
                self.state.assign(state)
        else:
            self.state.assign(state)

        self._init_forms()

    @staticmethod
    def default_solver_parameters():
        return NonlinearSolver.default_solver_parameters()

    def solve(self):
        r"""
        Solve the variational problem

        .. math::

           \delta W = 0

        """

        logger.debug("Solving variational problem")

        try:
            logger.debug("Try to solve")
            nliter, nlconv = self.solver.solve()

            if not nlconv:
                logger.debug("Failed")
                raise SolverDidNotConverge("Solver did not converge...")

        except RuntimeError as ex:
            logger.debug("Failed")
            logger.debug("Reintialize old state and raise exception")
            raise SolverDidNotConverge(ex) from ex
        else:
            logger.debug("Sucess")

        return nliter, nlconv

    def get_displacement(self, annotate=False):
        D = self.state_space.sub(0)
        V = D.collapse()

        fa = FunctionAssigner(V, D)
        u = Function(V, name="displacement")

        if has_dolfin_adjoint:
            fa.assign(u, self.state.split()[0], annotate=annotate)
        else:
            fa.assign(u, self.state.split()[0])

        return u

    def SecondPiolaStress(self):
        u, p = self.state.split(deepcopy=True)
        F = kinematics.DeformationGradient(u)
        return self.material.SecondPiolaStress(F, p)

    def FirstPiolaStress(self):
        u, p = self.state.split(deepcopy=True)
        F = kinematics.DeformationGradient(u)
        return self.material.FirstPiolaStress(F, p)

    def ChachyStress(self):
        u, p = self.state.split(deepcopy=True)
        F = kinematics.DeformationGradient(u)
        return self.material.CauchyStress(F, p)

    def GreenLagrangeStrain(self):
        u, p = self.state.split(deepcopy=True)
        F = kinematics.DeformationGradient(u)
        return kinematics.GreenLagrangeStrain(F)



#!/usr/bin/env python
from dolfin import det
from dolfin import grad as Grad
from dolfin import Identity
from dolfin import inner
from dolfin import inv
from dolfin import tr

from . import Constant
from .dolfin_utils import get_dimesion


# Second order identity tensor
def SecondOrderIdentity(F):
    """Return identity with same dimension as input"""
    dim = get_dimesion(F)
    return Identity(dim)


def DeformationGradient(u, isochoric=False):
    """Return deformation gradient from displacement."""
    Id = SecondOrderIdentity(u)
    F = Id + Grad(u)
    if isochoric:
        return IsochoricDeformationGradient(F)
    else:
        return F


def IsochoricDeformationGradient(F):
    """Return the isochoric deformation gradient"""
    J = Jacobian(F)
    dim = get_dimesion(F)
    return pow(J, -1.0 / float(dim)) * F


def Jacobian(F):
    """Determinant of the deformation gradient"""
    return det(F)


def EngineeringStrain(F, isochoric=False):
    """Equivalent of engineering strain"""
    Id = SecondOrderIdentity(F)
    if isochoric:
        return IsochoricDeformationGradient(F) - Id
    else:
        return F - Id


def GreenLagrangeStrain(F, isochoric=False):
    """Green-Lagrange strain tensor"""
    Id = SecondOrderIdentity(F)
    C = RightCauchyGreen(F, isochoric)
    return 0.5 * (C - Id)


def LeftCauchyGreen(F, isochoric=False):
    """Left Cauchy-Green tensor"""
    if isochoric:
        F_ = IsochoricDeformationGradient(F)
    else:
        F_ = F

    return F_ * F_.T


def RightCauchyGreen(F, isochoric=False):
    """Left Cauchy-Green tensor"""
    if isochoric:
        F_ = IsochoricDeformationGradient(F)
    else:
        F_ = F

    return F_.T * F_


def EulerAlmansiStrain(F, isochoric=False):
    """Euler-Almansi strain tensor"""
    Id = SecondOrderIdentity(F)
    b = LeftCauchyGreen(F, isochoric)
    return 0.5 * (Id - inv(b))


def I1(F, isochoric=False):
    C = RightCauchyGreen(F, isochoric)
    I1 = tr(C)
    return I1


def I2(F, isochoric=False):
    C = RightCauchyGreen(F, isochoric)
    return 0.5 * (I1(F) * I1(F) - tr(C * C))


def I3(F, isochoric=False):
    C = RightCauchyGreen(F, isochoric)
    return det(C)


def I4(F, a0=None, isochoric=False):
    if a0 is not None:
        C = RightCauchyGreen(F, isochoric)
        I4 = inner(C * a0, a0)
    else:
        I4 = Constant(0.0)
    return I4


def I5(F, a0, isochoric=False):
    if a0 is not None:
        C = RightCauchyGreen(F, isochoric)
        I5 = inner(C * a0, C * a0)
    else:
        I5 = Constant(0.0)
    return I5


def I6(F, b0):
    return I4(F, b0)


def I7(F, b0):
    return I5(F, b0)


def I8(F, a0, b0, isochoric=False):
    if a0 is None or b0 is None:
        I8 = Constant(0.0)
    else:
        F_ = IsochoricDeformationGradient(F) if isochoric else F
        I8 = inner(F_ * a0, F_ * b0)
    return I8


# Transforms #####
# Pull-back of a two-tensor from the current to the reference
# configuration
def PiolaTransform(A, F):
    J = Jacobian(F)
    B = J * A * inv(F).T
    return B


# Push-forward of a two-tensor from the reference to the current
# configuration
def InversePiolaTransform(A, F):
    J = Jacobian(F)
    B = (1 / J) * A * F.T
    return B



import dolfin

try:
    from dolfin_adjoint import Constant
except ImportError:
    from dolfin import Constant

from .. import kinematics
from .material_model import Material


class Guccione(Material):
    """
    Guccione material model.
    """

    name = "guccione"

    @staticmethod
    def default_parameters():
        p = {"C": 2.0, "bf": 8.0, "bt": 2.0, "bfs": 4.0}

        return p

    def SecondPiolaStress(self, F, p, *args, **kwargs):
        P = self.FirstPiolaStress(F, p)
        S = dolfin.inv(F) * P
        return S

    def is_isotropic(self):
        """
        Return True if the material is isotropic.
        """

        p = self.parameters
        return p["bt"] == 1.0 and p["bf"] == 1.0 and p["bfs"] == 1.0

    def strain_energy(self, F_):
        """
        UFL form of the strain energy.
        """

        params = self.parameters

        # Elastic part of deformation gradient
        F = self.Fe(F_)

        E = kinematics.GreenLagrangeStrain(F, isochoric=self.isochoric)

        CC = Constant(params["C"], name="C")

        e1 = self.f0
        e2 = self.s0
        e3 = self.n0

        if any(e is None for e in (e1, e2, e3)):
            msg = (
                "Need to provide the full orthotropic basis "
                "for the Guccione model got \ne1 = {e1}\n"
                "e2 = {e2}\ne3 = {e3}"
            ).format(e1=e1, e2=e2, e3=e3)
            raise ValueError(msg)

        if self.is_isotropic():
            # isotropic case
            Q = dolfin.inner(E, E)

        else:
            # fully anisotropic
            bt = Constant(params["bt"], name="bt")
            bf = Constant(params["bf"], name="bf")
            bfs = Constant(params["bfs"], name="bfs")

            E11, E12, E13 = (
                dolfin.inner(E * e1, e1),
                dolfin.inner(E * e1, e2),
                dolfin.inner(E * e1, e3),
            )
            _, E22, E23 = (
                dolfin.inner(E * e2, e1),
                dolfin.inner(E * e2, e2),
                dolfin.inner(E * e2, e3),
            )
            _, _, E33 = (
                dolfin.inner(E * e3, e1),
                dolfin.inner(E * e3, e2),
                dolfin.inner(E * e3, e3),
            )

            Q = (
                bf * E11**2
                + bt * (E22**2 + E33**2 + 2 * E23**2)
                + bfs * (2 * E12**2 + 2 * E13**2)
            )

        # passive strain energy
        Wpassive = CC / 2.0 * (dolfin.exp(Q) - 1)
        Wactive = self.Wactive(F, diff=0)

        return Wpassive + Wactive



import dolfin

from .. import kinematics
from .material_model import Material


class StVenantKirchhoff(Material):
    """
    Class for linear elastic material
    """

    name = "saint_venant_kirchhoff"

    @staticmethod
    def default_parameters():
        return {"mu": 300.0, "lmbda": 1.0}

    def strain_energy(self, F_):
        F = self.Fe(F_)
        E = kinematics.GreenLagrangeStrain(F, isochoric=self.isochoric)
        W = self.lmbda / 2 * (dolfin.tr(E) ** 2) + self.mu * dolfin.tr(E * E)

        # Active stress
        Wactive = self.Wactive(F, diff=0)

        return W + Wactive



import dolfin

from ..dolfin_utils import get_dimesion
from .material_model import Material


class LinearElastic(Material):
    """
    Class for linear elastic material
    """

    name = "linear_elastic"

    @staticmethod
    def default_parameters():
        return {"mu": 100.0, "lmbda": 1.0}

    def strain_energy(self, F_):
        F = self.Fe(F_)

        dim = get_dimesion(F)
        gradu = F - dolfin.Identity(dim)
        epsilon = 0.5 * (gradu + gradu.T)
        W = self.lmbda / 2 * (dolfin.tr(epsilon) ** 2) + self.mu * dolfin.tr(
            epsilon * epsilon,
        )

        # Active stress
        Wactive = self.Wactive(F, diff=0)

        return W + Wactive



from pathlib import Path

import dolfin
import h5py

from . import Function
from . import interpolate
from . import io_utils
from . import Mesh
from .utils import DOLFIN_VERSION_MAJOR
from .utils import getLogger
from .utils import mpi_comm_world

logger = getLogger(__name__)


def move(mesh, u, factor=1.0):
    """
    Move mesh according to some displacement times some factor
    """
    W = dolfin.VectorFunctionSpace(u.function_space().mesh(), "CG", 1)

    # Use interpolation for now. It is the only thing that makes sense
    u_int = interpolate(u, W)

    u0 = Function(W)
    # arr = factor * numpy_mpi.gather_vector(u_int.vector())
    # numpy_mpi.assign_to_vector(u0.vector(), arr)
    u0.vector()[:] = factor * u_int.vector()
    V = dolfin.VectorFunctionSpace(mesh, "CG", 1)
    U = Function(V)
    U.vector()[:] = u0.vector()
    # numpy_mpi.assign_to_vector(U.vector(), arr)

    dolfin.ALE.move(mesh, U)

    return u0


def get_geometric_dimension(mesh):
    try:
        return mesh.geometric_dimension()
    except AttributeError:
        return mesh.geometry().dim()


def load_geometry_from_h5(
    h5name,
    h5group="",
    include_sheets=True,
    comm=mpi_comm_world(),
):
    """Load geometry and other mesh data from
    a h5file to an object.
    If the file contains muliple fiber fields
    you can spefify the angles, and if the file
    contais sheets and cross-sheets this can also
    be included

    :param str h5name: Name of the h5file
    :param str h5group: The group within the file
    :param bool include_sheets: Include sheets and cross-sheets
    :returns: An object with geometry data
    :rtype: object

    """
    h5name = Path(h5name)

    logger.info("\nLoad mesh from h5")
    # Set default groups
    ggroup = f"{h5group}/geometry"
    mgroup = f"{ggroup}/mesh"
    lgroup = f"{h5group}/local basis functions"
    fgroup = f"{h5group}/microstructure/"

    if not h5name.is_file():
        raise IOError(f"File {h5name} does not exist")

    # Check that the given file contains
    # the geometry in the given h5group
    if not io_utils.check_h5group(h5name, mgroup, delete=False, comm=comm):
        msg = ("Warning!\nGroup: '{}' does not exist in file:" "\n{}").format(
            mgroup,
            h5name,
        )

        with h5py.File(h5name) as h:
            keys = h.keys()
        msg += f"\nPossible values for the h5group are {keys}"
        raise IOError(msg)

    # Create a dummy object for easy parsing
    class Geometry(object):
        pass

    geo = Geometry()

    with dolfin.HDF5File(comm, h5name.as_posix(), "r") as h5file:
        # Load mesh
        mesh = Mesh(comm)
        io_utils.read_h5file(h5file, mesh, mgroup, True)
        geo.mesh = mesh

        # Get mesh functions
        for dim, attr in enumerate(["vfun", "efun", "ffun", "cfun"]):
            if dim > get_geometric_dimension(mesh):
                setattr(geo, attr, None)
                continue

            dgroup = f"{ggroup}/mesh/meshfunction_{dim}"
            mf = dolfin.MeshFunction("size_t", mesh, dim, mesh.domains())

            if h5file.has_dataset(dgroup):
                io_utils.read_h5file(h5file, mf, dgroup)
            setattr(geo, attr, mf)

        load_local_basis(h5file, lgroup, mesh, geo)
        load_microstructure(h5file, fgroup, mesh, geo, include_sheets)

        # Load the boundary markers
        markers = load_markers(h5file, mesh, ggroup, dgroup)
        geo.markers = markers

        origmeshgroup = f"{h5group}/original_geometry"
        if h5file.has_dataset(origmeshgroup):
            original_mesh = Mesh(comm)
            io_utils.read_h5file(h5file, original_mesh, origmeshgroup, True)
            setattr(geo, "original_geometry", original_mesh)

    for attr in ["f0", "s0", "n0", "r0", "c0", "l0", "cfun", "vfun", "efun", "ffun"]:
        if not hasattr(geo, attr):
            setattr(geo, attr, None)

    return geo


def load_markers(h5file, mesh, ggroup, dgroup):
    try:
        markers = {}
        for dim in range(mesh.ufl_domain().topological_dimension() + 1):
            for key_str in ["domain", "meshfunction"]:
                dgroup = f"{ggroup}/mesh/{key_str}_{dim}"

                # If dataset is not present
                if not h5file.has_dataset(dgroup):
                    continue

                def get_attributes():
                    if DOLFIN_VERSION_MAJOR >= 2018:
                        return h5file.attributes(dgroup).list_attributes()
                    else:
                        return h5file.attributes(dgroup).str().strip().split(" ")

                for aname in get_attributes():
                    if aname.startswith("marker_name"):
                        name = aname.rsplit("marker_name_")[-1]
                        marker = h5file.attributes(dgroup)[f"marker_name_{name}"]
                        markers[name] = (int(marker), dim)

    except Exception as ex:
        logger.info("Unable to load makers")
        logger.info(ex)
        markers = {}

    return markers


def load_local_basis(h5file, lgroup, mesh, geo):
    if h5file.has_dataset(lgroup):
        # Get local bais functions
        local_basis_attrs = h5file.attributes(lgroup)
        lspace = local_basis_attrs["space"]
        family, order = lspace.split("_")

        namesstr = local_basis_attrs["names"]
        names = namesstr.split(":")

        if DOLFIN_VERSION_MAJOR > 1.6:
            elm = dolfin.VectorElement(
                family=family,
                cell=mesh.ufl_cell(),
                degree=int(order),
                quad_scheme="default",
            )
            V = dolfin.FunctionSpace(mesh, elm)
        else:
            V = dolfin.VectorFunctionSpace(mesh, family, int(order))

        for name in names:
            lb = Function(V, name=name)

            io_utils.read_h5file(h5file, lb, lgroup + f"/{name}")
            setattr(geo, name, lb)
    else:
        setattr(geo, "circumferential", None)
        setattr(geo, "radial", None)
        setattr(geo, "longitudinal", None)


def load_microstructure(h5file, fgroup, mesh, geo, include_sheets=True):
    if h5file.has_dataset(fgroup):
        # Get fibers
        fiber_attrs = h5file.attributes(fgroup)
        fspace = fiber_attrs["space"]
        if fspace is None:
            # Assume quadrature 4
            # family = "Quadrature"
            # order = 4
            family = "CG"
            order = 1
        else:
            family, order = fspace.split("_")

        namesstr = fiber_attrs["names"]
        if namesstr is None:
            names = ["fiber"]
        else:
            names = namesstr.split(":")

        # Check that these fibers exists
        for name in names:
            fsubgroup = fgroup + f"/{name}"
            if not h5file.has_dataset(fsubgroup):
                msg = ("H5File does not have dataset {}").format(fsubgroup)
                logger.warning(msg)

        if DOLFIN_VERSION_MAJOR > 1.6:
            elm = dolfin.VectorElement(
                family=family,
                cell=mesh.ufl_cell(),
                degree=int(order),
                quad_scheme="default",
            )
            V = dolfin.FunctionSpace(mesh, elm)
        else:
            V = dolfin.VectorFunctionSpace(mesh, family, int(order))

        attrs = ["f0", "s0", "n0"]
        for i, name in enumerate(names):
            func = Function(V, name=name)
            fsubgroup = fgroup + f"/{name}"

            io_utils.read_h5file(h5file, func, fsubgroup)

            setattr(geo, attrs[i], func)


def save_meshfunctions(h5file, h5group, mesh, meshfunctions, comm):
    for dim in range(get_geometric_dimension(mesh) + 1):
        if meshfunctions is not None and dim in meshfunctions:
            mf = meshfunctions[dim]
        else:
            mf = dolfin.MeshFunction("size_t", mesh, dim, mesh.domains())

        save_mf = dolfin.MPI.max(comm, len(set(mf.array()))) > 1

        if save_mf:
            dgroup = f"{ggroup(h5group)}/mesh/meshfunction_{dim}"
            h5file.write(mf, dgroup)


def save_markers(h5file, h5group, markers):
    if markers is None:
        return
    # Save the boundary markers
    for name, (marker, dim) in markers.items():
        for key_str in ["domain", "meshfunction"]:
            dgroup = f"{ggroup(h5group)}/mesh/{key_str}_{dim}"

            if h5file.has_dataset(dgroup):
                aname = f"marker_name_{name}"
                h5file.attributes(dgroup)[aname] = marker


def ggroup(h5group):
    return f"{h5group}/geometry"


def mgroup(h5group):
    return f"{ggroup(h5group)}/mesh"


def save_geometry_to_h5(
    mesh,
    h5name,
    h5group="",
    markers=None,
    fields=None,
    local_basis=None,
    meshfunctions=None,
    comm=None,
    other_functions=None,
    other_attributes=None,
    overwrite_file=False,
    overwrite_group=True,
):
    """
    Save geometry and other geometrical functions to a HDF file.

    Parameters
    ----------
    mesh : :class:`dolfin.mesh`
        The mesh
    h5name : str
        Path to the file
    h5group : str
        Folder within the file. Default is "" which means in
        the top folder.
    markers : dict
        A dictionary with markers. See `get_markers`.
    fields : list
        A list of functions for the microstructure
    local_basis : list
        A list of functions for the crl basis
    meshfunctions : dict
        A dictionary with keys being the dimensions the the values
        beeing the meshfunctions.
    comm : :class:`dolfin.MPI`
        MPI communicator
    other_functions : dict
        Dictionary with other functions you want to save
    other_attributes: dict
        Dictionary with other attributes you want to save
    overwrite_file : bool
        If true, and the file exists, the file will be overwritten (default: False)
    overwrite_group : bool
        If true and h5group exist, the group will be overwritten.

    """

    logger.info("\nSave mesh to h5")
    assert isinstance(mesh, dolfin.Mesh)
    if comm is None:
        comm = mesh.mpi_comm()

    h5name = Path(h5name)
    file_mode = "a" if h5name.is_file() and not overwrite_file else "w"

    # IF we should append the file but overwrite the group we need to
    # check that the group does not exist. If so we need to open it in
    # h5py and delete it.
    if file_mode == "a" and overwrite_group:
        if h5group != "":
            io_utils.check_h5group(h5name, h5group, delete=True, comm=comm)
        else:
            h5name.unlink()
            file_mode = "w"

    with dolfin.HDF5File(comm, h5name.as_posix(), file_mode) as h5file:
        # Save mesh
        h5file.write(mesh, mgroup(h5group))
        save_meshfunctions(h5file, h5group, mesh, meshfunctions, comm)
        save_markers(h5file, h5group, markers)
        save_local_basis(h5file, h5group, local_basis)
        save_fields(h5file, h5group, fields)
        save_other_functions(h5file, h5group, other_functions)
        save_other_attributes(h5file, h5group, other_attributes)
    logger.info(f"Geometry saved to {h5name}")


def save_fields(h5file, h5group, fields):
    """Save mictrostructure to h5file

    Parameters
    ----------
    h5file : dolfin.HDF5File
        The file to write to
    h5group : str
        The folder inside the file to write to
    fields : list
        List of dolfin functions to write
    """
    if fields is None:
        return
    fgroup = f"{h5group}/microstructure"
    names = []
    for field in fields:
        try:
            label = field.label() if field.label().rfind("a Function") == -1 else ""
        except AttributeError:
            label = field.name()
        name = "_".join(filter(None, [str(field), label]))
        fsubgroup = f"{fgroup}/{name}"
        h5file.write(field, fsubgroup)
        h5file.attributes(fsubgroup)["name"] = field.name()
        names.append(name)

    elm = field.function_space().ufl_element()
    family, degree = elm.family(), elm.degree()
    fspace = f"{family}_{degree}"
    h5file.attributes(fgroup)["space"] = fspace
    h5file.attributes(fgroup)["names"] = ":".join(names)


def save_other_functions(h5file, h5group, other_functions):
    if other_functions is None:
        return
    for k, fun in other_functions.items():
        fungroup = "/".join([h5group, k])
        h5file.write(fun, fungroup)

        if isinstance(fun, dolfin.Function):
            elm = fun.function_space().ufl_element()
            family, degree, vsize = elm.family(), elm.degree(), elm.value_size()
            fspace = f"{family}_{degree}"
            h5file.attributes(fungroup)["space"] = fspace
            h5file.attributes(fungroup)["value_size"] = vsize


def save_other_attributes(h5file, h5group, other_attributes):
    if other_attributes is None:
        return
    for k, v in other_attributes.iteritems():
        if isinstance(v, str) and isinstance(k, str):
            h5file.attributes(h5group)[k] = v
        else:
            logger.warning(f"Invalid attribute {k} = {v}")


def save_local_basis(h5file, h5group, local_basis):
    """Save local basis functions

    Parameters
    ----------
    h5file : dolfin.HDF5File
        The file to write to
    h5group : str
        Folder inside file to store the local basis
    local_basis : list
        List of dolfin functions with local basis functions
    """
    if local_basis is None:
        return
    lgroup = f"{h5group}/local basis functions"
    names = []
    for basis in local_basis:
        h5file.write(basis, lgroup + f"/{basis.name()}")
        names.append(basis.name())

    elm = basis.function_space().ufl_element()
    family, degree = elm.family(), elm.degree()
    lspace = f"{family}_{degree}"
    h5file.attributes(lgroup)["space"] = lspace
    h5file.attributes(lgroup)["names"] = ":".join(names)



import dolfin as df
import pytest

from pulse import kinematics

A = 1.1
B = 1 / A
C = 1


@pytest.fixture
def F_2D_Real():
    mesh = df.UnitSquareMesh(2, 2)
    V = df.TensorFunctionSpace(mesh, "R", 0)
    F = df.Function(V)
    F.assign(df.Constant([[A, 0], [0.0, B]]))
    return F


@pytest.fixture
def F_2D_CG1():
    mesh = df.UnitSquareMesh(2, 2)
    V = df.TensorFunctionSpace(mesh, "CG", 1)
    F = df.Function(V)
    F.assign(df.Constant([[A, 0], [0.0, B]]))
    return F


@pytest.fixture
def F_3D_Real():
    mesh = df.UnitCubeMesh(2, 2, 2)
    V = df.TensorFunctionSpace(mesh, "R", 0)
    F = df.Function(V)
    F.assign(df.Constant([[A, 0, 0], [0.0, B, 0], [0, 0, C]]))
    return F


@pytest.fixture
def F_3D_CG1():
    mesh = df.UnitCubeMesh(2, 2, 2)
    V = df.TensorFunctionSpace(mesh, "CG", 1)
    F = df.Function(V)
    F.assign(df.Constant([[A, 0, 0], [0.0, B, 0], [0, 0, C]]))
    return F


@pytest.fixture
def F_dict(F_2D_Real, F_2D_CG1, F_3D_Real, F_3D_CG1):
    return {
        "F_2D_Real": F_2D_Real,
        "F_2D_CG1": F_2D_CG1,
        "F_3D_Real": F_3D_Real,
        "F_3D_CG1": F_3D_CG1,
    }


@pytest.mark.parametrize(
    "F_str, dim",
    [("F_2D_Real", 2), ("F_2D_CG1", 2), ("F_3D_Real", 3), ("F_3D_CG1", 3)],
)
def test_SecondOderIdetity(F_dict, F_str, dim):
    F = F_dict[F_str]
    Id = kinematics.SecondOrderIdentity(F)
    assert Id == df.Identity(dim)


@pytest.mark.parametrize("F_str", ["F_2D_Real", "F_2D_CG1", "F_3D_Real", "F_3D_CG1"])
def test_EngineeringStrain(F_dict, F_str):
    F = F_dict[F_str]
    E = kinematics.EngineeringStrain(F)
    assert abs(df.assemble((E[0, 0]) * df.dx) - (A - 1)) < 1e-12
    assert abs(df.assemble((E[1, 1]) * df.dx) - (B - 1)) < 1e-12
    assert abs(df.assemble((E[1, 0]) * df.dx)) < 1e-12
    assert abs(df.assemble((E[0, 1]) * df.dx)) < 1e-12


@pytest.mark.parametrize("F_str", ["F_2D_Real", "F_2D_CG1", "F_3D_Real", "F_3D_CG1"])
def test_GreenLagrangeStrain(F_dict, F_str):
    F = F_dict[F_str]
    E = kinematics.GreenLagrangeStrain(F)
    assert abs(df.assemble((E[0, 0]) * df.dx) - (0.5 * (A**2 - 1))) < 1e-12
    assert abs(df.assemble((E[1, 1]) * df.dx) - (0.5 * (B**2 - 1))) < 1e-12
    assert abs(df.assemble((E[1, 0]) * df.dx)) < 1e-12
    assert abs(df.assemble((E[0, 1]) * df.dx)) < 1e-12


@pytest.mark.parametrize("F_str", ["F_2D_Real", "F_2D_CG1", "F_3D_Real", "F_3D_CG1"])
def test_I1(F_dict, F_str):
    F = F_dict[F_str]
    c = 0 if "2D" in F_str else C
    # Trace of F**2
    assert abs(df.assemble(kinematics.I1(F) * df.dx) - (A**2 + B**2 + c**2)) < 1e-12


@pytest.mark.parametrize("F_str", ["F_2D_Real", "F_2D_CG1", "F_3D_Real", "F_3D_CG1"])
def test_I2(F_dict, F_str):
    F = F_dict[F_str]
    c = 0 if "2D" in F_str else C

    assert (
        abs(
            df.assemble(kinematics.I2(F) * df.dx)
            - 0.5 * ((A**2 + B**2 + c**2) ** 2 - (A**4 + B**4 + c**4)),
        )
        < 1e-12
    )


@pytest.mark.parametrize("F_str", ["F_2D_Real", "F_2D_CG1", "F_3D_Real", "F_3D_CG1"])
def test_I3(F_dict, F_str):
    F = F_dict[F_str]
    c = 1 if "2D" in F_str else C

    assert abs(df.assemble(kinematics.I3(F) * df.dx) - A * B * c) < 1e-12


@pytest.mark.parametrize("F_str", ["F_2D_Real", "F_2D_CG1", "F_3D_Real", "F_3D_CG1"])
def test_I4x(F_dict, F_str):
    F = F_dict[F_str]

    if "2D" in F_str:
        x = df.as_vector([1, 0])
    else:
        x = df.as_vector([1, 0, 0])

    assert abs(df.assemble(kinematics.I4(F, x) * df.dx) - A**2) < 1e-12


@pytest.mark.parametrize("F_str", ["F_2D_Real", "F_2D_CG1", "F_3D_Real", "F_3D_CG1"])
def test_I4y(F_dict, F_str):
    F = F_dict[F_str]

    if "2D" in F_str:
        y = df.as_vector([0, 1])
    else:
        y = df.as_vector([0, 1, 0])

    assert abs(df.assemble(kinematics.I4(F, y) * df.dx) - (B) ** 2) < 1e-12


@pytest.mark.parametrize("F_str", ["F_2D_Real", "F_2D_CG1", "F_3D_Real", "F_3D_CG1"])
def test_I5x(F_dict, F_str):
    F = F_dict[F_str]

    if "2D" in F_str:
        x = df.as_vector([1, 0])
    else:
        x = df.as_vector([1, 0, 0])

    assert abs(df.assemble(kinematics.I5(F, x) * df.dx) - A**4) < 1e-12


@pytest.mark.parametrize("F_str", ["F_2D_Real", "F_2D_CG1", "F_3D_Real", "F_3D_CG1"])
def test_I5y(F_dict, F_str):
    F = F_dict[F_str]

    if "2D" in F_str:
        y = df.as_vector([0, 1])
    else:
        y = df.as_vector([0, 1, 0])

    assert abs(df.assemble(kinematics.I5(F, y) * df.dx) - B**4) < 1e-12


def test_I6():
    """I6 is identical to I4"""
    pass


def test_I7():
    """I7 is identical to I5"""
    pass


@pytest.mark.parametrize("F_str", ["F_2D_Real", "F_2D_CG1", "F_3D_Real", "F_3D_CG1"])
def test_I8xy(F_dict, F_str):
    F = F_dict[F_str]

    if "2D" in F_str:
        x = df.as_vector([1, 0])
        y = df.as_vector([0, 1])
    else:
        x = df.as_vector([1, 0, 0])
        y = df.as_vector([0, 1, 0])

    assert abs(df.assemble(kinematics.I8(F, x, y) * df.dx) - (A**2 * 0 + 0 * B)) < 1e-12



from .. import kinematics
from ..dolfin_utils import get_dimesion
from .material_model import Material


class NeoHookean(Material):
    """
    Class for Neo Hookean material
    """

    name = "neo_hookean"

    @staticmethod
    def default_parameters():
        return {"mu": 15.0}

    def W_1(self, I_1, diff=0, dim=3, *args, **kwargs):
        mu = self.mu

        if diff == 0:
            return 0.5 * mu * (I_1 - dim)
        elif diff == 1:
            return 0.5 * mu
        elif diff == 2:
            return 0

    def strain_energy(self, F):
        r"""
        Strain-energy density function.

        .. math::

           \mathcal{W} = \mathcal{W}_1 + \mathcal{W}_{4f}
           + \mathcal{W}_{\mathrm{active}}

        where

        .. math::

           \mathcal{W}_{\mathrm{active}} =
           \begin{cases}
             0 & \text{if acitve strain} \\
             \gamma I_{4f} & \text{if active stress}
           \end{cases}


        :param F: Deformation gradient
        :type F: :py:class:`dolfin.Function`

        """

        # Invariants
        I1 = kinematics.I1(self.Fe(F), isochoric=self.isochoric)

        # Active stress
        Wactive = self.Wactive(F, diff=0)

        dim = get_dimesion(F)
        W1 = self.W_1(I1, diff=0, dim=dim)

        W = W1 + Wactive

        return W



import os
import sys
from pathlib import Path
from unittest import mock

import pytest

try:
    import mshr  # noqa: F401
except ImportError:
    has_mshr = False
else:
    has_mshr = True

# There are some problems with some plotting backends in containers.
# Therefore we set the backend here, and execute the files in stead
# of spawning subprocesses.
import matplotlib

matplotlib.use("agg")

here = Path(__file__).parent
demodir = here.parent.joinpath("demo")


demos = [
    (f, root)
    for root, dirname, files in os.walk(demodir)
    for f in files
    if Path(f).suffix == ".py"
]


@pytest.mark.parametrize("filename, root", demos)
def test_demo(filename, root):
    if os.path.basename(root) in ["closed_loop", "optimal_control", "custom_geometry"]:
        return
    if os.path.basename(root) == "creating_geometries" and not has_mshr:
        return
    os.chdir(root)
    # Add the current folder to sys.path so that
    # python finds the relevant modules
    py_filename = Path(filename)
    sys.path.append(root)
    # Execute file
    with mock.patch("pulse.mechanicsproblem.MechanicsProblem.solve") as solve_mock:
        solve_mock.return_value = (1, True)  # (niter, nconv)
        exec(open(py_filename).read(), globals())
    # Remove the current folder from the sys.path
    sys.path.pop()



#!/usr/bin/env python
import logging
from typing import Optional

import dolfin
import numpy as np


class MPIFilt(logging.Filter):
    def filter(self, record):
        if dolfin.MPI.rank(mpi_comm_world()) == 0:
            return 1
        else:
            return 0


mpi_filt = MPIFilt()


def getLogger(name):
    import daiquiri

    logger = daiquiri.getLogger(name)
    logger.logger.addFilter(mpi_filt)
    return logger


logger = getLogger(__name__)


def get_dolfin_version():
    if dolfin.__version__.startswith("20"):
        # Year based versioning
        return float(dolfin.__version__.split(".")[0])
    else:
        return float(".".join(dolfin.__version__.split(".")[:2]))


try:
    DOLFIN_VERSION_MAJOR = get_dolfin_version()
except AttributeError:
    # Just assume the lastest one
    DOLFIN_VERSION_MAJOR = 2019.0


class Annotation(object):
    """
    Object holding global annotation for dolfin-adjoint
    """

    def __init__(self):
        self.annotate = False

    @property
    def annotate(self):
        return self._annotate

    @annotate.setter
    def annotate(self, annotate=False):
        """
        Set global annotation for dolfin-adjoint.
        Default False
        """
        if "adjoint" in dolfin.parameters:
            dolfin.parameters["adjoint"]["stop_annotating"] = not annotate
        else:
            try:
                from pyadjoint.tape import _stop_annotating
            except ImportError:
                # Dolfin-adjoint is most likely not installed
                pass
            else:
                _stop_annotating = not annotate  # noqa: F841,F811

        # Update local variable
        self._annotate = annotate


try:
    annotation: Optional[Annotation] = Annotation()
except Exception:
    annotation = None


class Enlisted(tuple):
    pass


def enlist(x, force_enlist=False):
    if isinstance(x, Enlisted):
        return x
    elif isinstance(x, (list, tuple, np.ndarray)):
        if force_enlist:
            return Enlisted([x])
        else:
            return Enlisted(x)
    else:
        return Enlisted([x])


def delist(x):
    if isinstance(x, Enlisted):
        if len(x) == 1:
            return x[0]
        else:
            return x
    else:
        return x


def mpi_comm_world():
    if DOLFIN_VERSION_MAJOR >= 2018:
        return dolfin.MPI.comm_world
    else:
        return dolfin.mpi_comm_world()


def value_size(obj):
    try:
        return obj.value_size()
    except AttributeError:
        value_shape = obj.value_shape()
        if len(value_shape) == 0:
            return 1
        else:
            return [0]


# Dummy object
class Object(object):
    pass


def get_lv_marker(geometry):
    for key in ["ENDO", "ENDO_LV", "endo", "endo_lv"]:
        if key in geometry.markers:
            return geometry.markers[key][0]
    raise KeyError("Geometry is missing marker for LV ENDO")


class UnableToChangePressureExeption(Exception):
    pass



#!/usr/bin/env python
# c) 2001-2017 Simula Research Laboratory ALL RIGHTS RESERVED
# Authors: Henrik Finsberg
# END-USER LICENSE AGREEMENT
# PLEASE READ THIS DOCUMENT CAREFULLY. By installing or using this
# software you agree with the terms and conditions of this license
# agreement. If you do not accept the terms of this license agreement
# you may not install or use this software.
# Permission to use, copy, modify and distribute any part of this
# software for non-profit educational and research purposes, without
# fee, and without a written agreement is hereby granted, provided
# that the above copyright notice, and this license agreement in its
# entirety appear in all copies. Those desiring to use this software
# for commercial purposes should contact Simula Research Laboratory AS:
# post@simula.no
#
# IN NO EVENT SHALL SIMULA RESEARCH LABORATORY BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE
# "PULSE-ADJOINT" EVEN IF SIMULA RESEARCH LABORATORY HAS BEEN ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE. THE SOFTWARE PROVIDED HEREIN IS
# ON AN "AS IS" BASIS, AND SIMULA RESEARCH LABORATORY HAS NO OBLIGATION
# TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
# SIMULA RESEARCH LABORATORY MAKES NO REPRESENTATIONS AND EXTENDS NO
# WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESSED, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
from abc import ABC
from abc import abstractmethod
from abc import abstractstaticmethod
from typing import Optional
from typing import Union

import dolfin

try:
    import ufl_legacy as ufl
except ImportError:
    import ufl

from . import active_model as _active_model
from .. import Constant
from .. import Function
from .. import kinematics
from .. import numpy_mpi
from .. import project
from ..dolfin_utils import RegionalParameter
from ..dolfin_utils import update_function
from ..geometry import HeartGeometry


def compressibility(model, *args, **kwargs):
    if model == "incompressible":
        return incompressible(*args, **kwargs)


def incompressible(p, J):
    return -p * (J - 1.0)


_t = _active_model.ActiveStressModels.transversally


class Material(ABC):
    """
    Initialize material model

    Parameters
    ----------

    f0 : :py:class`dolfin.Function`
        Fiber field
    gamma : :py:class`dolfin.Function`
        Activation parameter
    params : dict
        Material parameters
    active_model : str
        Active model - active strain or active stress
    s0 : :py:class`dolfin.Function`
        Sheets
    n0 : :py:class`dolfin.Function`
        Sheet - normal
    T_ref : float
        Scale factor for activation parameter (default = 1.0)
    dev_iso_split : bool
        Decouple deformation into deviatoric and isochoric deformations
    eta : float
        Fraction of transverse active tesion for active stress formulation.
        0 = active only along fiber, 1 = equal forces in all directions
        (default=0.0).
    """

    def __init__(
        self,
        activation: Optional[ufl.Coefficient] = None,
        parameters=None,
        active_model: Union[
            _active_model.ActiveModel,
            _active_model.ActiveModels,
        ] = _active_model.ActiveModels.active_strain,
        T_ref: Optional[float] = None,
        eta: float = 0.0,
        isochoric: bool = True,
        compressible_model="incompressible",
        geometry=None,
        f0=None,
        s0=None,
        n0=None,
        active_isotropy: _active_model.ActiveStressModels = _t,
        *args,
        **kwargs,
    ):
        # Parameters
        self.parameters = self.__class__.default_parameters()
        if parameters is not None:
            self.parameters.update(parameters)
        self._set_parameter_attrs(geometry)

        active = active_model
        if not isinstance(active_model, _active_model.ActiveModel):
            active = _active_model.ActiveModel(
                activation=activation,
                f0=f0,
                s0=s0,
                n0=n0,
                model=active_model,
                active_isotropy=active_isotropy,
                eta=eta,
                T_ref=T_ref,
                isochoric=isochoric,
            )
        self.active = active
        self.compressible_model = compressible_model

    def _set_parameter_attrs(self, geometry: Optional[HeartGeometry] = None):
        for k, v in self.parameters.items():
            if isinstance(v, (float, int)):
                setattr(self, k, Constant(v, name=k))

            elif isinstance(v, RegionalParameter):
                if geometry is not None:
                    v_new = RegionalParameter(geometry.cfun)
                    numpy_mpi.assign_to_vector(
                        v_new.vector(),
                        numpy_mpi.gather_vector(v.vector()),
                    )
                    v = v_new

                ind_space = v.proj_space
                setattr(self, k, Function(ind_space, name=k))
                mat = getattr(self, k)
                matfun = v.function
                mat.assign(project(matfun, ind_space))

            else:
                if geometry is not None and v.ufl_element().cell() is not None:
                    v_new = update_function(geometry.mesh, v)
                    v = v_new

                setattr(self, k, v)

    @abstractmethod
    def strain_energy(self, F):
        pass

    @staticmethod
    @abstractstaticmethod
    def default_parameters():
        pass

    def update_geometry(self, geometry):
        # Loop over the possible attributes containing
        # a domain and update it
        for m in ("f0", "s0", "n0"):
            setattr(self, m, getattr(geometry, m))

        self._set_parameter_attrs(geometry)

        activation_element = self.activation.ufl_element()
        if activation_element.cell() is not None:
            self.active._activation = update_function(geometry.mesh, self.activation)

    def copy(self, geometry=None):
        if geometry is not None:
            f0, s0, n0 = geometry.f0, geometry.s0, geometry.n0
        else:
            f0, s0, n0 = self.f0, self.s0, self.n0

        return self.__class__(
            activation=self.activation,
            parameters=self.parameters,
            active_model=self.active_model,
            geometry=geometry,
            f0=f0,
            s0=s0,
            n0=n0,
            T_ref=float(self.T_ref),
            eta=float(self.eta),
            isochoric=self.isochoric,
        )

    @property
    def name(self):
        return "generic_material"

    def __repr__(self):
        return (
            "{self.__class__.__name__}({self.parameters}, "
            "{self.active_model}, "
            "{self.compressible_model})"
        ).format(self=self)

    def compressibility(self, p, J):
        return compressibility(self.compressible_model, p, J)

    @property
    def Fa(self):
        return self.active.Fa

    def Fe(self, F):
        return self.active.Fe(F)

    def Wactive(self, F=None, diff=0):
        """Active stress energy"""
        return self.active.Wactive(F=F, diff=diff)

    @property
    def isochoric(self):
        return self.active.isochoric

    @property
    def active_model(self):
        return self.active.model

    @property
    def material_model(self):
        return self._model

    @property
    def eta(self):
        return self.active.eta

    @property
    def T_ref(self):
        return self.active.T_ref

    @property
    def f0(self):
        return self.active.f0

    @f0.setter
    def f0(self, f0):
        self.active.f0 = f0

    @property
    def s0(self):
        return self.active.s0

    @s0.setter
    def s0(self, s0):
        self.active.s0 = s0

    @property
    def n0(self):
        return self.active.n0

    @n0.setter
    def n0(self, n0):
        self.active.n0 = n0

    @property
    def activation_field(self):
        """
        Return the activation field.
        If regional, this will return a piecewise
        constant function (DG_0)
        """
        return self.active.activation_field

    @property
    def activation(self):
        """
        Return the activation paramter.
        If regional, this will return one parameter
        for each segment.
        """
        return self.active.activation

    def CauchyStress(self, F, p=None, deviatoric=False):
        F = dolfin.variable(F)

        # First Piola Kirchoff
        if deviatoric:
            p = None

        P = self.FirstPiolaStress(F, p)

        # Cauchy stress
        T = kinematics.InversePiolaTransform(P, F)

        return T

    def SecondPiolaStress(self, F, p=None, deviatoric=False, *args, **kwargs):
        # First Piola Kirchoff
        if deviatoric:
            p = None

        P = self.FirstPiolaStress(F, p)

        S = dolfin.inv(F) * P * dolfin.inv(F).T

        return S

    def FirstPiolaStress(self, F, p=None, *args, **kwargs):
        F = dolfin.variable(F)

        # First Piola Kirchoff
        psi_iso = self.strain_energy(F)
        P = dolfin.diff(psi_iso, F)

        if p is not None:
            J = dolfin.variable(kinematics.Jacobian(F))
            psi_vol = self.compressibility(p, J)
            # PiolaTransform
            P_vol = J * dolfin.diff(psi_vol, J) * dolfin.inv(F).T

            P += P_vol

        return P



#!/usr/bin/env python
import operator as op

import dolfin
import numpy as np

from . import Constant
from . import Constants
from . import Function
from . import Functions
from . import has_dolfin_adjoint
from . import numpy_mpi
from .dolfin_utils import get_constant
from .mechanicsproblem import SolverDidNotConverge
from .utils import delist
from .utils import enlist
from .utils import Enlisted
from .utils import getLogger
from .utils import value_size

logger = getLogger(__name__)


def copy(f, deepcopy=True, name="copied_function"):
    """
    Copy a function. This is to ease the integration
    with dolfin adjoint where copied fuctions are annotated.
    """

    if isinstance(f, Functions):
        if has_dolfin_adjoint:
            try:
                return f.copy(deepcopy=deepcopy, name=name)
            except TypeError:
                return f.copy(deepcopy=deepcopy)
        else:
            return f.copy(deepcopy=deepcopy)
    elif isinstance(f, dolfin.Constant):
        if f.value_size() > 1:
            return dolfin.Constant(f.values(), name=name)
        return dolfin.Constant(f, name=name)
    elif isinstance(f, Constant):
        return Constant(f, name=name)
    elif isinstance(f, (float, int)):
        return f
    elif isinstance(f, Enlisted):
        lst = []
        for fi in f:
            lst.append(copy(fi))
        return enlist(lst)

    elif isinstance(f, (list, tuple)):
        lst = []
        for fi in f:
            lst.append(copy(fi))
        return tuple(lst)
    else:
        return f


def constant2float(const):
    """
    Convert a :class:`dolfin.Constant` to float
    """
    try:
        c = float(const)
    except Exception:
        const = get_constant(const)
        try:
            c = float(const)
        except TypeError:
            try:
                c = np.zeros(len(const))
                const.eval(c, c)
            except Exception as ex:
                logger.warning(ex, exc_info=True)
                return const
    return c


def get_delta(new_control, c0, c1):
    """
    Get extrapolation parameter used in the continuation step.
    """
    if isinstance(c0, Constants):
        c0 = constant2float(c0)
        c1 = constant2float(c1)
        new_control = constant2float(new_control)

    if isinstance(new_control, (int, float)):
        delta = (new_control - c0) / float(c1 - c0)

    elif isinstance(new_control, (tuple, np.ndarray, list)):
        c0 = [constant2float(c) for c in c0]
        c1 = [constant2float(c) for c in c1]
        new_control = [constant2float(c) for c in new_control]
        delta = (new_control[0] - c0[0]) / float(c1[0] - c0[0])

    elif isinstance(new_control, (dolfin.GenericVector, dolfin.Vector)):
        new_control_arr = numpy_mpi.gather_vector(new_control)
        c0_arr = numpy_mpi.gather_vector(c0)
        c1_arr = numpy_mpi.gather_vector(c1)

        delta = (
            np.subtract(new_control_arr, c0_arr).mean()
            / np.subtract(c1_arr, c0_arr).mean()
        )

    elif isinstance(new_control, (dolfin.Function, Function)):
        new_control_arr = numpy_mpi.gather_vector(
            new_control.vector(),
            new_control.function_space().dim(),
        )
        c0_arr = numpy_mpi.gather_vector(c0.vector(), c0.function_space().dim())
        c1_arr = numpy_mpi.gather_vector(c1.vector(), c1.function_space().dim())

        delta = (
            np.subtract(new_control_arr, c0_arr).mean()
            / np.subtract(c1_arr, c0_arr).mean()
        )

    else:
        msg = ("Unexpected type for new_crontrol in get_delta" "Got {}").format(
            type(new_control),
        )
        raise TypeError(msg)

    return squeeze(delta)


def np2str(c_, fmt="{:.3f}"):
    c = delist(c_)
    if isinstance(c, (np.ndarray, list, tuple)):
        return ", ".join([f"{ci:.3f}" for ci in c])
    return f"{c:.3f}"


def print_control(cs, msg):
    controls = [constant2float(c) for c in cs]

    if len(controls) > 3:
        msg += ("\n\tMin:{:.2f}\tMean:{:.2f}\tMax:{:.2f}" "").format(
            np.min(controls),
            np.mean(controls),
            np.max(controls),
        )
    else:
        cs = []
        for c in controls:
            if hasattr(c, "__len__"):
                msg += print_control(c, msg)
            else:
                cs.append(c)
        if cs:
            msg += ",".join([np2str(c) for c in controls])
    return msg


def get_diff(current, target):
    """
    Get difference between current and target value
    """
    if isinstance(target, (Function, dolfin.Function)):
        diff = target.vector() - current.vector()

    elif isinstance(target, Constants):
        diff = np.subtract(constant2float(target), constant2float(current))
    elif isinstance(target, (tuple, list)):
        diff = np.subtract(
            [constant2float(t) for t in target],
            [constant2float(c) for c in current],
        )
    else:
        try:
            diff = np.subtract(target, current)
        except Exception as ex:
            logger.error(ex)
            raise ValueError(
                ("Unable to compute diff with type {}" "").format(type(current)),
            )

    return squeeze(diff)


def squeeze(x):
    try:
        y = np.squeeze(x)
    except Exception:
        return x
    else:
        try:
            shape = np.shape(y)
        except Exception:
            return y
        else:
            if len(shape) == 0:
                return float(y)
            else:
                return y


def get_initial_step(current, target, nsteps=5):
    """
    Estimate the step size needed to step from current to target
    in `nsteps`.
    """

    diff = get_diff(current, target)
    if isinstance(diff, dolfin.GenericVector):
        step = Function(current.function_space())
        step.vector().axpy(1.0 / float(nsteps), diff)

    else:
        step = diff / float(nsteps)

    logger.debug(
        ("Intial number of steps: {} with step size {}" "").format(nsteps, step),
    )
    return step


def step_too_large(current, target, step):
    """
    Check if `current + step` exceeds `target`
    """
    if isinstance(target, (dolfin.Function, Function)):
        target = numpy_mpi.gather_vector(target.vector(), target.function_space().dim())
    elif isinstance(target, Constants):
        target = constant2float(target)
    target = squeeze(target)

    if isinstance(current, (dolfin.Function, Function)):
        current = numpy_mpi.gather_vector(
            current.vector(),
            current.function_space().dim(),
        )
    elif isinstance(current, Constants):
        current = constant2float(current)
    current = squeeze(current)

    if isinstance(step, (dolfin.Function, Function)):
        step = numpy_mpi.gather_vector(step.vector(), step.function_space().dim())
    elif isinstance(step, Constants):
        step = constant2float(step)
    step = squeeze(step)

    if not hasattr(target, "__len__"):
        cond = current < target
        if hasattr(cond, "__len__"):
            cond = np.any(cond)

        comp = op.gt if cond else op.lt
        return np.any(comp(current + step, target))
    else:
        too_large = []
        for c, t, s in zip(current, target, step):
            too_large.append(step_too_large(c, t, s))

    return np.any(too_large)


def iterate(
    problem,
    control,
    target,
    continuation=True,
    max_adapt_iter=8,
    adapt_step=True,
    old_states=None,
    old_controls=None,
    max_nr_crash=20,
    max_iters=40,
    initial_number_of_steps=5,
    reinit_each_step=False,
):
    """
    Using the given problem, iterate control to given target.

    Arguments
    ---------
    problem : pulse.MechanicsProblem
        The problem
    control : dolfin.Function or dolfin.Constant
        The control
    target: dolfin.Function, dolfin.Constant, tuple or float
        The target value. Typically a float if target is LVP, a tuple
        if target is (LVP, RVP) and a function if target is gamma.
    continuation: bool
        Apply continuation for better guess for newton problem
        Note: Replay test seems to fail when continuation is True,
        but taylor test passes
    max_adapt_iter: int
        If number of iterations is less than this number and adapt_step=True,
        then adapt control step. Default: 8
    adapt_step: bool
        Adapt / increase step size when sucessful iterations are achevied.
    old_states: list
        List of old controls to help speed in the continuation
    reinit_each_step : bool
        If True reinitialize form at each step.
    """

    with Iterator(
        problem=problem,
        control=control,
        target=target,
        continuation=continuation,
        max_adapt_iter=max_adapt_iter,
        adapt_step=adapt_step,
        old_states=old_states,
        old_controls=old_controls,
        max_nr_crash=max_nr_crash,
        max_iters=max_iters,
        initial_number_of_steps=initial_number_of_steps,
        reinit_each_step=reinit_each_step,
    ) as iterator:
        res = iterator.solve()

    return res


class Iterator(object):
    """
    Iterator
    """

    _control_types = (Function, dolfin.Function, Constant, dolfin.Constant)

    def __init__(
        self,
        problem,
        control,
        target,
        old_states=None,
        old_controls=None,
        **params,
    ):
        self.parameters = Iterator.default_parameters()
        self.parameters.update(params)

        self.old_controls = () if old_controls is None else old_controls
        self.old_states = () if old_states is None else old_states
        self.problem = problem

        self._check_control(control)
        self._check_target(target)

        self.control_values = [
            copy(delist(self.control), deepcopy=True, name="previous control"),
        ]
        self.prev_states = [
            copy(self.problem.state, deepcopy=True, name="previous state"),
        ]

        self.step = get_initial_step(
            self.control,
            self.target,
            self.parameters["initial_number_of_steps"],
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        pass

    @staticmethod
    def default_parameters():
        return dict(
            continuation=True,
            max_adapt_iter=8,
            adapt_step=True,
            max_nr_crash=20,
            max_iters=40,
            initial_number_of_steps=5,
            reinit_each_step=False,
        )

    @property
    def step(self):
        return self._step

    @step.setter
    def step(self, step):
        s = squeeze(step)
        if hasattr(s, "__len__") and len(s) != len(self.control):
            if len(self.control) == 1:
                # Then probably the control is a function with
                # dimension higher than one and we have squeezed to much
                s = enlist(s, force_enlist=True)
            else:
                msg = (
                    "Step is of lenght {} while the lenght " "of the control is {}"
                ).format(len(s), len(self.control))
                raise ValueError(msg)
        else:
            s = enlist(s)

        self._step = s

    def solve(self):
        self.ncrashes = 0
        self.niters = 0
        # print(type(self.control))
        control_name = self.control[0].name()
        logger.info(f"Iterating to target control ({control_name})...")
        msg = f"Current control: {control_name} = "
        logger.info(print_control(self.control, msg))
        msg = "Target: "
        logger.info(print_control(self.target, msg))
        while not self.target_reached():
            self.niters += 1
            if (
                self.ncrashes > self.parameters["max_nr_crash"]
                or self.niters > self.parameters["max_iters"]
            ):
                self.problem.reinit(self.prev_states[0])
                self.assign_control(enlist(self.control_values[0]))

                raise SolverDidNotConverge

            prev_state = self.prev_states[-1]
            prev_control = enlist(self.control_values[-1])

            # Check if we are close
            if step_too_large(
                delist(prev_control),
                delist(self.target),
                delist(self.step),
            ):
                self.change_step_for_final_iteration(delist(prev_control))

            self.increment_control()

            if self.parameters["continuation"]:
                self.continuation_step()

            logger.debug("Try new control")
            self.print_control()

            try:
                nliter, nlconv = self.problem.solve()

            except SolverDidNotConverge as ex:
                logger.debug(ex)
                logger.debug("\nNOT CONVERGING")
                logger.debug("Reduce control step")
                self.ncrashes += 1
                self.assign_control(prev_control)
                # Assign old state
                logger.debug("Assign old state")
                self.problem.state.vector().zero()
                self.problem.reinit(prev_state)

                self.change_step_size(0.5)

            else:
                self.ncrashes = 0
                logger.debug("\nSUCCESFULL STEP:")

                if not self.target_reached():
                    if (
                        nliter < self.parameters["max_adapt_iter"]
                        and self.parameters["adapt_step"]
                    ):
                        self.change_step_size(1.5)
                        msg = f"Adapt step size. New step size: {np2str(self.step)}"
                        # print_control(enlist(self.step), msg)
                        logger.debug(msg)

                self.control_values.append(
                    delist(copy(self.control, deepcopy=True, name="Previous control")),
                )

                self.prev_states.append(
                    copy(self.problem.state, deepcopy=True, name="Previous state"),
                )
        return self.prev_states, self.control_values

    def change_step_size(self, factor):
        step = delist(self.step)
        if isinstance(step, (list, tuple)):
            self.step = [factor * s for s in step]
        else:
            self.step = factor * step

    def print_control(self):
        msg = "Current control: "
        msg = print_control(self.control, msg)
        logger.debug(msg)

    def continuation_step(self):
        first_step = len(self.prev_states) < 2
        if first_step:
            return

        c0, c1 = self.control_values[-2:]
        s0, s1 = self.prev_states[-2:]

        delta = get_delta(delist(self.control), c0, c1)

        if has_dolfin_adjoint:
            w = dolfin.Function(self.problem.state.function_space())

            w.vector().zero()
            w.vector().axpy(1.0 - delta, s0.vector())
            w.vector().axpy(delta, s1.vector())
            self.problem.reinit(w, annotate=True)
        else:
            self.problem.state.vector().zero()
            self.problem.state.vector().axpy(1.0 - delta, s0.vector())
            self.problem.state.vector().axpy(delta, s1.vector())

    def increment_control(self):
        for c, s in zip(self.control, self.step):
            if isinstance(c, (dolfin.Function, Function)):
                c.vector()[:] += s
            else:
                c.assign(Constant(constant2float(c) + s))

        if self.parameters["reinit_each_step"]:
            self.problem._init_forms()

    def assign_control(self, new_control):
        """
        Assign a new value to the control
        """
        for c, n in zip(self.control, new_control):
            try:
                c.assign(n)
            except TypeError:
                c.assign(Constant(n))

    def change_step_for_final_iteration(self, prev_control):
        """Change step size so that target is
        reached in the next iteration
        """
        logger.debug("Change step size for final iteration")

        target = delist(self.target)
        prev_control = delist(prev_control)

        if isinstance(target, (dolfin.Function, Function)):
            step = Function(target.function_space())
            step.vector().axpy(1.0, target.vector())
            step.vector().axpy(-1.0, prev_control.vector())
        elif isinstance(target, (list, np.ndarray, tuple)):
            if isinstance(prev_control, (dolfin.Function, Function)):
                prev = numpy_mpi.gather_vector(
                    prev_control.vector(),
                    prev_control.function_space().dim(),
                )
            else:
                prev = prev_control

            step = np.array(
                [constant2float(t) - constant2float(c) for (t, c) in zip(target, prev)],
            )
        elif isinstance(target, (dolfin.Constant, Constant)):
            step = constant2float(target) - constant2float(prev_control)
        else:
            step = target - prev_control

        self.step = step

    def _check_target(self, target):
        target = enlist(target)

        targets = []
        for tar in target:
            try:
                t = get_constant(tar)
            except TypeError:
                msg = ("Unable to convert target for type {} " "to a constant").format(
                    type(target),
                )
                raise TypeError(msg)
            targets.append(t)

        self.target = Enlisted(targets)
        t0 = self.target[0]
        msg = (
            "Unsuppoert shape of control and target. "
            "Can only handle single arrays or multiple scalars."
        )
        for t in self.target[1:]:
            if value_size(t0) > 1 or value_size(t) > 1:
                raise (ValueError(msg))
        logger.debug(f"Target: {[constant2float(t) for t in self.target]}")

    def _check_control(self, control):
        control = enlist(control)
        # Control has to be either a function or
        # a constant
        for c in control:
            msg = ("Expected control parameters to be of type {}, " "got {}").format(
                self._control_types,
                type(c),
            )
            assert isinstance(c, self._control_types), msg

        self.control = control
        logger.debug(f"Control: {[constant2float(c) for c in self.control]}")

    @property
    def ncontrols(self):
        """Number of controls"""
        return len(self.control)

    def target_reached(self):
        """Check if control and target are the same"""
        diff = get_diff(self.control, self.target)

        if isinstance(diff, dolfin.GenericVector):
            diff.abs()
            max_diff = diff.max()

        else:
            max_diff = np.max(abs(diff))

        reached = max_diff < 1e-6
        if reached:
            logger.debug("Check target reached: YES!")
        else:
            logger.debug("Check target reached: NO")
            logger.debug(f"Maximum difference: {max_diff:.3e}")

        return reached



# Material models
# Active models
from .active_model import ActiveModel
from .active_model import ActiveModels
from .guccione import Guccione
from .holzapfelogden import HolzapfelOgden
from .linearelastic import LinearElastic
from .material_model import Material
from .neohookean import NeoHookean
from .stvenantkirchhoff import StVenantKirchhoff

material_models = (
    HolzapfelOgden,
    Guccione,
    NeoHookean,
    LinearElastic,
    StVenantKirchhoff,
)

material_model_names = [m.name for m in material_models]


def get_material_model(material_model):
    for model in material_models:
        if model.name == material_model:
            return model

    raise ValueError(
        ("Material model {} does not exist. " "Please use one of {}").format(
            material_model,
            material_model_names,
        ),
    )


__all__ = [
    "Material",
    "HolzapfelOgden",
    "Guccione",
    "NeoHookean",
    "LinearElastic",
    "StVenantKirchhoff",
    "ActiveModel",
    "ActiveModels",
]



#!/usr/bin/env python
"""These functions are copied from
cbcpost https://bitbucket.org/simula_cbc/cbcpost
"""
import dolfin
import numpy as np
from dolfin import MPI

from .utils import DOLFIN_VERSION_MAJOR
from .utils import mpi_comm_world


def gather_vector(u, size=None):
    return u


def compile_extension_module(cpp_code, **kwargs):
    if DOLFIN_VERSION_MAJOR >= 2018:
        headers = kwargs.get("additional_system_headers", [])
        headers = [
            "#include <Eigen/Core>",
            "#include <pybind11/pybind11.h>",
            # "using Array = Eigen::Ref<const Eigen::ArrayXi>;;"] +\
            "using Array = Eigen::ArrayXi;",
        ] + ["#include <" + h + ">" for h in headers if h != ""]
        cpp_code = "\n".join(headers) + cpp_code
        return dolfin.compile_cpp_code(cpp_code)
    else:
        return dolfin.compile_extension_module(cpp_code, **kwargs)


def broadcast(array, from_process):
    "Broadcast array to all processes"
    if not hasattr(broadcast, "cpp_module"):
        cpp_code = """

        namespace dolfin {
            std::vector<double> broadcast(
                const MPI_Comm mpi_comm,
                const Array<double>& inarray, int from_process)
            {
                int this_process = dolfin::MPI::rank(mpi_comm);
                std::vector<double> outvector(inarray.size());

                if(this_process == from_process) {
                    for(int i=0; i<inarray.size(); i++)
                    {
                        outvector[i] = inarray[i];
                    }
                }
                dolfin::MPI::barrier(mpi_comm);
                dolfin::MPI::broadcast(mpi_comm, outvector, from_process);

                return outvector;
            }
        }
        """
        if DOLFIN_VERSION_MAJOR >= 2018:
            cpp_code += """

            PYBIND11_MODULE(SIGNATURE, m)
            {
            m.def("broadcast", &dolfin::broadcast);
            }
            """
        cpp_module = compile_extension_module(
            cpp_code,
            additional_system_headers=["dolfin/common/MPI.h"],
        )

        broadcast.cpp_module = cpp_module

    cpp_module = broadcast.cpp_module

    if MPI.rank(mpi_comm_world()) == from_process:
        array = np.array(array, dtype=np.float)
        shape = array.shape
        shape = np.array(shape, dtype=np.float_)
    else:
        array = np.array([], dtype=np.float)
        shape = np.array([], dtype=np.float_)

    shape = cpp_module.broadcast(mpi_comm_world(), shape, from_process)
    array = array.flatten()

    out_array = cpp_module.broadcast(mpi_comm_world(), array, from_process)

    if len(shape) > 1:
        out_array = out_array.reshape(*shape)

    return out_array


def gather(array, on_process=0, flatten=False):
    "Gather array from all processes on a single process"
    if not hasattr(gather, "cpp_module"):
        cpp_code = """
        namespace dolfin {
            std::vector<double> gather(
                const MPI_Comm mpi_comm,
                const Array<double>& inarray, int on_process)
            {
                int this_process = dolfin::MPI::rank(mpi_comm);

                std::vector<double> outvector(
                    dolfin::MPI::size(mpi_comm)*dolfin::MPI::sum(mpi_comm, inarray.size())
                );
                std::vector<double> invector(inarray.size());

                for(int i=0; i<inarray.size(); i++)
                {
                    invector[i] = inarray[i];
                }

                dolfin::MPI::gather(mpi_comm, invector, outvector, on_process);
                return outvector;
            }
        }
        """
        if DOLFIN_VERSION_MAJOR >= 2018:
            cpp_code += """

            PYBIND11_MODULE(SIGNATURE, m)
            {
            m.def("gather", &dolfin::gather);
            }
            """
        gather.cpp_module = compile_extension_module(
            cpp_code,
            additional_system_headers=["dolfin/common/MPI.h"],
        )

    cpp_module = gather.cpp_module
    array = np.array(array, dtype=np.float)
    out_array = cpp_module.gather(mpi_comm_world(), array, on_process)

    if flatten:
        return out_array

    dist = distribution(len(array))
    cumsum = [0] + [sum(dist[: i + 1]) for i in range(len(dist))]
    out_array = [[out_array[cumsum[i] : cumsum[i + 1]]] for i in range(len(cumsum) - 1)]

    return out_array


def distribution(number):
    "Get distribution of number on all processes"
    if not hasattr(distribution, "cpp_module"):
        cpp_code = """
        namespace dolfin {
            std::vector<unsigned int> distribution(const MPI_Comm mpi_comm, int number)
            {
                // Variables to help in synchronization
                int num_processes = dolfin::MPI::size(mpi_comm);
                int this_process = dolfin::MPI::rank(mpi_comm);

                std::vector<uint> distribution(num_processes);

                for(uint i=0; i<num_processes; i++) {
                    if(i==this_process) {
                        distribution[i] = number;
                    }
                    dolfin::MPI::barrier(mpi_comm);
                    dolfin::MPI::broadcast(mpi_comm, distribution, i);
                }
                return distribution;
          }
        }
        """
        if DOLFIN_VERSION_MAJOR >= 2018:
            cpp_code += """

            PYBIND11_MODULE(SIGNATURE, m)
            {
            m.def("distribution", &dolfin::distribution);
            }
            """
        distribution.cpp_module = compile_extension_module(
            cpp_code,
            additional_system_headers=["dolfin/common/MPI.h"],
        )

    cpp_module = distribution.cpp_module
    return cpp_module.distribution(mpi_comm_world(), number)


def gather_broadcast(arr):
    # try:
    #     dtype = arr.dtype
    # except AttributeError:
    #     dtype = np.float

    # arr = gather(arr, flatten=True)
    # arr = broadcast(arr, 0)

    # return arr.astype(dtype)
    return arr


def assign_to_vector(v, a):
    """
    Assign the value of the array a to the dolfin vector v
    """
    lr = v.local_range()
    v[:] = a[lr[0] : lr[1]]


def mpi_print(mess):
    if mpi_comm_world().rank == 0:
        print(mess)
